import argparse
import datetime
import logging
import os
from typing import Sequence

import apache_beam as beam
import metview
import numpy as np
import pandas as pd
import xarray as xr
import xarray_beam as xbeam
from apache_beam.options.pipeline_options import PipelineOptions


def attribute_fix(ds):
    """Needed to fix a low-level bug in ecCodes.

    Sometimes, shortNames get overloaded in ecCodes's table.
    To eliminate ambiguity in their string matching, we
    force ecCodes to make use of the paramId, which is a
    consistent source-of-truth.
    """
    for var in ds:
        ds[var].attrs.pop("GRIB_cfName", None)
        ds[var].attrs.pop("GRIB_cfVarName", None)
        ds[var].attrs.pop("GRIB_shortName", None)

    return ds


def grid_attribute_fix(ds, names_to_fix, reference_name):
    """Avoid an error raised by MetView when trying to regrid certain 2D variables."""

    attrs_to_replace = [
        "GRIB_NV",
        "GRIB_gridDefinitionDescription",
        "GRIB_gridType",
        "GRIB_latitudeOfFirstGridPointInDegrees",
        "GRIB_latitudeOfLastGridPointInDegrees",
        "GRIB_longitudeOfFirstGridPointInDegrees",
        "GRIB_longitudeOfLastGridPointInDegrees",
    ]
    for name in names_to_fix:
        new_attrs = {k: ds[reference_name].attrs[k] for k in attrs_to_replace}
        ds[name] = ds[name].assign_attrs(new_attrs)
    return ds


TEMPLATE_PATH = "gs://vcm-ml-scratch/oliwm/2024-04-22-era5-regrid-template.zarr"

GRID_DOCS_URL = "https://confluence.ecmwf.int/display/OIFS/4.3+OpenIFS%3A+Horizontal+Resolution+and+Configurations"  # noqa: E501
DEFAULT_OUTPUT_GRID = "F90"  # 1° regular Gaussian grid. See GRID_DOCS_URL linked above.
TIME_STEP = 6  # in same units as resolution of time coordinate of data (i.e. hours)
GRAVITY = 9.80665  # value used in metview according to https://metview.readthedocs.io/en/latest/metview/macro/functions/fieldset.html#id0 # noqa: E501

URL_GOOGLE_ARCO_ERA5 = "gs://gcp-public-data-arco-era5/co"
URL_WIND_TEMP = f"{URL_GOOGLE_ARCO_ERA5}/model-level-wind.zarr-v2"
URL_SURFACE = f"{URL_GOOGLE_ARCO_ERA5}/single-level-surface.zarr-v2"
URL_SURFACE_REANALYSIS = f"{URL_GOOGLE_ARCO_ERA5}/single-level-reanalysis.zarr-v2"
URL_MOISTURE = f"{URL_GOOGLE_ARCO_ERA5}/model-level-moisture.zarr-v2"
URL_GOOGLE_LATLON = (
    "gs://gcp-public-data-arco-era5/ar/full_37-1h-0p25deg-chunk-1.zarr-v3"
)
# following dataset was manually generated in https://github.com/ai2cm/explore/blob/master/oliwm/2024-07-31-generate-ERA5-co2.ipynb # noqa: E501
URL_CO2 = "gs://vcm-ml-raw-flexible-retention/2024-11-11-co2-annual-mean-for-era5.zarr"

URL_NCAR_ERA5 = (
    "gs://vcm-ml-intermediate/2024-05-17-era5-025deg-2D-variables-from-NCAR-as-zarr"  # noqa: E501
)
URL_INVARIANT = f"{URL_NCAR_ERA5}/e5.oper.invariant.zarr"
URL_SURFACE_ANALYSIS_LATLON = f"{URL_NCAR_ERA5}/e5.oper.an.sfc.zarr"
URL_MEAN_FLUX = f"{URL_NCAR_ERA5}/e5.oper.fc.sfc.meanflux.zarr"

WIND_TEMP = "wind_temp"
SURFACE = "surface"
SURFACE_REANALYSIS = "surface_reanalysis"
MOISTURE = "moisture"
INVARIANT = "invariant_latlon"
SURFACE_ANALYSIS_LATLON = "surface_analysis_latlon"
MEAN_FLUX = "mean_flux_latlon"
GOOGLE_LATLON = "google_latlon"

URLS = {
    WIND_TEMP: URL_WIND_TEMP,
    SURFACE: URL_SURFACE,
    MOISTURE: URL_MOISTURE,
    SURFACE_REANALYSIS: URL_SURFACE_REANALYSIS,
    INVARIANT: URL_INVARIANT,
    SURFACE_ANALYSIS_LATLON: URL_SURFACE_ANALYSIS_LATLON,
    MEAN_FLUX: URL_MEAN_FLUX,
    GOOGLE_LATLON: URL_GOOGLE_LATLON,
}

# these versions of the zarr datasets have variables with the necessary attrs for
# Metview to work (just stripping the '-v2' from the ends of the URLs)
URLS_WITH_REQUIRED_ATTRS = {
    WIND_TEMP: URL_WIND_TEMP[:-3],
    SURFACE: URL_SURFACE[:-3],
    MOISTURE: URL_MOISTURE[:-3],
    SURFACE_REANALYSIS: URL_SURFACE_REANALYSIS[:-3],
}

VARIABLE_NAMES = {
    WIND_TEMP: ["d", "vo", "t"],
    SURFACE: ["lnsp"],
    MOISTURE: ["q", "clwc", "ciwc", "crwc", "cswc"],
    SURFACE_REANALYSIS: ["skt", "t2m", "u10", "v10", "d2m"],
    INVARIANT: ["LSM", "Z"],
    SURFACE_ANALYSIS_LATLON: ["CI", "SWVL1", "SWVL2", "SWVL3", "SWVL4"],
    MEAN_FLUX: [
        "MER",
        "MSDWLWRF",
        "MSDWSWRF",
        "MSLHF",
        "MSNLWRF",
        "MSNSWRF",
        "MSR",
        "MSSHF",
        "MTDWSWRF",
        "MTNLWRF",
        "MTNSWRF",
        "MTPR",
        "MVIMD",
    ],
    GOOGLE_LATLON: [
        "specific_humidity",
        "temperature",
        "u_component_of_wind",
        "v_component_of_wind",
        "geopotential",
        "total_column_water_vapour",
    ],
}

VALUE_COORD_NAME = {
    WIND_TEMP: "spharm_index",
    SURFACE: "spharm_index",
    MOISTURE: "reduced_gg_index",
    SURFACE_REANALYSIS: "reduced_gg_index",
}

SCALAR_COORDS_TO_DROP = [
    "hybrid",
    "valid_time",
    "step",
    "number",
    "heightAboveGround",
    "surface",
    "entireAtmosphere",
    "depthBelowLandLayer",
    "level",
]

DESIRED_ATTRS = {
    "DSWRFtoa": {"long_name": "Downward SW radiative flux at TOA", "units": "W/m**2"},
    "USWRFtoa": {"long_name": "Upward SW radiative flux at TOA", "units": "W/m**2"},
    "ULWRFtoa": {"long_name": "Upward LW radiative flux at TOA", "units": "W/m**2"},
    "DSWRFsfc": {
        "long_name": "Downward SW radiative flux at surface",
        "units": "W/m**2",
    },
    "USWRFsfc": {"long_name": "Upward SW radiative flux at surface", "units": "W/m**2"},
    "DLWRFsfc": {
        "long_name": "Downward LW radiative flux at surface",
        "units": "W/m**2",
    },
    "ULWRFsfc": {"long_name": "Upward LW radiative flux at surface", "units": "W/m**2"},
    "LHTFLsfc": {"long_name": "Latent heat flux", "units": "W/m**2"},
    "SHTFLsfc": {"long_name": "Sensible heat flux", "units": "W/m**2"},
    "PRATEsfc": {"long_name": "Surface precipitation rate", "units": "kg/m**2/s"},
    "tendency_of_total_water_path_due_to_advection": {
        "long_name": "Tendency of total water path due to advection",
        "units": "kg/m**2/s",
    },
    "HGTsfc": {"long_name": "Topography height", "units": "m"},
    "land_fraction": {"long_name": "land fraction"},
    "sea_ice_fraction": {"long_name": "sea ice fraction"},
    "ocean_fraction": {"long_name": "ocean fraction"},
    "PRESsfc": {"long_name": "Surface pressure", "units": "Pa"},
    "TMP2m": {"long_name": "2m air temperature", "units": "K"},
    "Q2m": {"long_name": "2m specific humidity", "units": "kg/kg"},
    "DPT2m": {"long_name": "2m dewpoint temperature", "units": "K"},
    "UGRD10m": {"long_name": "10m U component of wind", "units": "m/s"},
    "VGRD10m": {"long_name": "10m V component of wind", "units": "m/s"},
}

# The input data is on the L137 ECMWF grid. See
# https://confluence.ecmwf.int/display/UDOC/L137+model+level+definitions.
# The indices below are chosen for closest alignment with the ACE vertical
# grid defined in Table 2 of https://arxiv.org/pdf/2310.02074.pdf except
# that the uppermost layer uses the higher model top of ECMWF model.
N_INPUT_LAYERS = 137  # this is the number of full layers, not interfaces
DEFAULT_OUTPUT_LAYER_INDICES = [0, 48, 67, 79, 90, 100, 109, 119, 137]

OUTPUT_PRESSURE_LEVELS = [850, 500, 200]  # additionally save these pressure levels
OUTPUT_PRESSURE_LEVELS_GEOPOTENTIAL = [1000, 850, 700, 500, 300, 250, 200]

RENAME_Q_PRES = {f"specific_humidity_{p}": f"Q{p}" for p in OUTPUT_PRESSURE_LEVELS}
RENAME_T_PRES = {f"temperature_{p}": f"TMP{p}" for p in OUTPUT_PRESSURE_LEVELS}
RENAME_U_PRES = {f"u_component_of_wind_{p}": f"UGRD{p}" for p in OUTPUT_PRESSURE_LEVELS}
RENAME_V_PRES = {f"v_component_of_wind_{p}": f"VGRD{p}" for p in OUTPUT_PRESSURE_LEVELS}
RENAME_Z_PRES = {
    f"geopotential_{p}": f"h{p}" for p in OUTPUT_PRESSURE_LEVELS_GEOPOTENTIAL
}
RENAME_PRESSURE_LEVEL = {
    **RENAME_Q_PRES,
    **RENAME_T_PRES,
    **RENAME_U_PRES,
    **RENAME_V_PRES,
    **RENAME_Z_PRES,
}


def _get_native_rename_dict(n_output_layers):
    rename_q = {f"q_{i}": f"specific_total_water_{i}" for i in range(n_output_layers)}
    rename_t = {f"t_{i}": f"air_temperature_{i}" for i in range(n_output_layers)}
    rename_u = {f"u_{i}": f"eastward_wind_{i}" for i in range(n_output_layers)}
    rename_v = {f"v_{i}": f"northward_wind_{i}" for i in range(n_output_layers)}
    rename_etc = {
        "skt": "surface_temperature",
        "t2m": "TMP2m",
        "u10": "UGRD10m",
        "v10": "VGRD10m",
        "d2m": "DPT2m",
    }
    rename_native = {
        **rename_q,
        **rename_t,
        **rename_u,
        **rename_v,
        **rename_etc,
    }
    return rename_native


def _open_zarr(key, sel_indices):
    ds = xr.open_zarr(URLS[key], chunks=None)
    ds = ds[VARIABLE_NAMES[key]]
    if key in VALUE_COORD_NAME:
        ds = ds.rename({"values": VALUE_COORD_NAME[key]})
    # xarray does not raise an error if selecting beyond bounds of time coord
    # so we manually check here that all desired data is available.
    if key != INVARIANT:
        dims = list(sel_indices[key])
        for dim in dims:
            desired_start = sel_indices[key][dim].start
            desired_stop = sel_indices[key][dim].stop
            ds_start = pd.Timestamp(ds[dim].min().values.item())
            ds_stop = pd.Timestamp(ds[dim].max().values.item())
            assert desired_start >= ds_start, f"{key} dataset {dim} start out of bounds"
            assert desired_stop <= ds_stop, f"{key} dataset {dim} stop out of bounds"
    ds = ds.sel(**sel_indices[key])
    if key == INVARIANT:
        ds = ds.drop_vars("time")
    if key in URLS_WITH_REQUIRED_ATTRS:
        tmp = xr.open_zarr(URLS_WITH_REQUIRED_ATTRS[key])
        for name in ds:
            if name in tmp:
                ds[name].attrs = tmp[name].attrs
    ds = attribute_fix(ds)
    return ds


def open_native_datasets(indices) -> xr.Dataset:
    native_dataset_keys = [WIND_TEMP, SURFACE, SURFACE_REANALYSIS, MOISTURE]
    datasets = [_open_zarr(k, indices) for k in native_dataset_keys]
    merged = xr.merge(datasets, compat="override", join="override")
    merged = grid_attribute_fix(merged, VARIABLE_NAMES[SURFACE_REANALYSIS], "q")
    return merged


def open_google_latlon_dataset(indices) -> xr.Dataset:
    ds = _open_zarr(GOOGLE_LATLON, indices)
    return ds


def open_meanflux_dataset(indices) -> xr.Dataset:
    ds = _open_zarr(MEAN_FLUX, indices)
    return ds


def open_quarter_degree_datasets(indices) -> xr.Dataset:
    sfc = _open_zarr(SURFACE_ANALYSIS_LATLON, indices)
    invariant = _open_zarr(INVARIANT, indices)
    return sfc, invariant


def open_co2_dataset(start_time, end_time) -> xr.Dataset:
    co2 = xr.open_zarr(URL_CO2, chunks=None)
    ds_start = pd.Timestamp(co2.time.values[0])
    ds_stop = pd.Timestamp(co2.time.values[-1])
    assert start_time >= ds_start, f"CO2 dataset time start out of bounds"
    assert end_time <= ds_stop, f"CO2 dataset time stop out of bounds"
    co2 = co2.sel(time=slice(start_time, end_time))
    co2 = co2.load()
    return co2


def _to_dataset(fs: metview.Fieldset) -> xr.Dataset:
    return fs.to_dataset().load()


def _to_dataarray(fs: metview.Fieldset, name: str) -> xr.DataArray:
    return fs.to_dataset()[name].load()


def _delete_fs(fs: metview.Fieldset):
    # manually delete the temporary grib file that MetView creates
    path = fs.url()
    if os.path.exists(path):
        os.remove(path)
    del fs


def _saturation_vapor_pressure(t: xr.DataArray) -> xr.DataArray:
    """https://metview.readthedocs.io/en/latest/api/functions/saturation_vapour_pressure.html"""  # noqa: E501
    a1 = 611.21
    a2 = 273.16
    a3 = 17.502
    a4 = 32.19
    return a1 * np.exp(a3 * (t - a2) / (t - a4))


def _specific_humidity_from_dewpoint(
    dewpoint: xr.DataArray, pressure: xr.DataArray
) -> xr.DataArray:
    """https://metview.readthedocs.io/en/latest/api/functions/specific_humidity_from_dewpoint.html
    https://metview.readthedocs.io/en/latest/api/functions/vapour_pressure.html"""
    ewsat = _saturation_vapor_pressure(dewpoint)
    eps = 0.621981
    result = eps * ewsat / (pressure - (1 - eps) * ewsat)
    result.attrs["units"] = "kg / kg"
    result.attrs["long_name"] = "Specific humidity"
    return result


def _to_geopotential_height(geopotential: xr.DataArray) -> xr.DataArray:
    output = geopotential / GRAVITY
    output.attrs["long_name"] = "Geopotential height"
    output.attrs["units"] = "m"
    output.attrs["standard_name"] = "geopotential_height"
    return output


def _process_pressure_level_data(ds: xr.Dataset, output_grid: str) -> xr.Dataset:
    """Select pressure levels from 0.25° pressure level dataset."""
    # convert to 2D fields at desired pressure levels
    select_levels = xr.Dataset()
    for name in ds.data_vars:
        if "level" in ds[name].dims:
            if name == "geopotential":
                pressure_levels = OUTPUT_PRESSURE_LEVELS_GEOPOTENTIAL
            else:
                pressure_levels = OUTPUT_PRESSURE_LEVELS
            for pressure in pressure_levels:
                logging.info(f"Selecting {name} at {pressure} hPa")
                out_name = f"{name}_{pressure}"
                select_levels[out_name] = ds[name].sel(level=pressure)
                if name == "geopotential":
                    select_levels[out_name] = _to_geopotential_height(
                        select_levels[out_name]
                    )
                select_levels[out_name].attrs["long_name"] += f" at {pressure} hPa"
        else:
            select_levels[name] = ds[name]

    regridded = _regrid_quarter_degree(select_levels, output_grid)
    regridded = regridded.rename(RENAME_PRESSURE_LEVEL)
    # coordinates will be written by template, so drop here to avoid possible conflicts
    regridded = regridded.drop_vars(["latitude", "longitude"])

    return regridded


def process_pressure_level_data(key, ds, output_grid=DEFAULT_OUTPUT_GRID):
    output = _process_pressure_level_data(ds, output_grid)
    new_key = key.replace(
        offsets={"time": key.offsets["time"], "latitude": 0, "longitude": 0},
        vars=frozenset(output.keys()),
    )
    return new_key, output


def _process_native_data(
    ds: xr.Dataset, output_grid: str, output_layer_indices: Sequence[int]
) -> xr.Dataset:
    n_output_layers = len(output_layer_indices) - 1
    rename_dict = _get_native_rename_dict(n_output_layers)

    xr.set_options(keep_attrs=True)
    # singleton time dimension interferes with metview
    ds = ds.squeeze()

    logging.info("Starting calculation of horizontal winds on output grid")
    vort_div_fieldset = metview.dataset_to_fieldset(ds[["d", "vo"]].load())
    winds_spharm = metview.uvwind(data=vort_div_fieldset)
    winds_gg = metview.regrid(data=winds_spharm, grid=output_grid, truncation="none")

    logging.info("Starting calculation of total specific water")
    total_specific_water = ds.q + ds.clwc + ds.ciwc + ds.crwc + ds.cswc
    total_specific_water.attrs = ds.q.attrs
    ds["total_specific_water"] = total_specific_water

    logging.info("Starting regrid of T, q and log surface pressure to output grid")
    names = ["total_specific_water", "t", "lnsp"]
    # next step converts total_specific_water to q because of attributes
    fieldset = metview.dataset_to_fieldset(ds[names].load())
    fieldset_gg = metview.regrid(data=fieldset, grid=output_grid, truncation="none")

    logging.info("Merging with wind fieldset")
    fieldset_gg = fieldset_gg.merge(winds_gg)

    logging.info("Computing vertical integrals of T, q, u and v")
    output = xr.Dataset()
    thicknesses_fs = metview.unithickness(fieldset_gg.select(shortName="lnsp"))
    thicknesses = _to_dataarray(thicknesses_fs, "pres")
    for short_name in ["q", "t", "u", "v"]:
        variable = _to_dataarray(fieldset_gg.select(shortName=short_name), short_name)
        for output_index in range(n_output_layers):
            logging.info(
                f"Computing vertical integral of {short_name} "
                f"for output layer {output_index}."
            )

            fine_levels = slice(
                output_layer_indices[output_index],
                output_layer_indices[output_index + 1],
            )
            coarse_level_thicknesses = thicknesses.isel(hybrid=fine_levels)
            total_thickness = coarse_level_thicknesses.sum("hybrid")
            coarse_level_variable = variable.isel(hybrid=fine_levels)
            pressure_weighted = coarse_level_variable * coarse_level_thicknesses
            integrated = pressure_weighted.sum("hybrid") / total_thickness

            out_name = f"{short_name}_{output_index}"
            output[out_name] = integrated
            if short_name == "q":
                output[out_name].attrs["long_name"] = "Specific total water"
                output[out_name].attrs["standard_name"] = "specific_total_water"
            output[out_name].attrs["long_name"] += f" level-{output_index}"
    del thicknesses
    del variable

    logging.info("Computing surface pressure and inserting into output dataset")
    output["PRESsfc"] = np.exp(
        _to_dataarray(fieldset_gg.select(shortName="lnsp"), "lnsp")
    )

    logging.info("Regridding additional 2D fields and inserting into output dataset")
    # keys are names in zarr dataset, values are what metview change the names to
    grib_names = {
        "skt": "skt",
        "u10": "u",
        "v10": "v",
        "t2m": "t",
        "d2m": "dpt",
    }
    fieldset_2d = metview.dataset_to_fieldset(ds[list(grib_names)].load())
    fieldset_2d_gg = metview.regrid(data=fieldset_2d, grid=output_grid)
    dataset_2d_gg = _to_dataset(fieldset_2d_gg)
    for name in grib_names:
        output[name] = dataset_2d_gg[grib_names[name]]

    # insert 2m specific humidity
    output["Q2m"] = _specific_humidity_from_dewpoint(output["d2m"], output["PRESsfc"])

    output = _adjust_latlon(output)

    output = output.rename(rename_dict)
    for name, attrs in DESIRED_ATTRS.items():
        if name in output:
            output[name] = output[name].assign_attrs(**attrs)

    output = output.drop_vars(SCALAR_COORDS_TO_DROP, errors="ignore")

    # coordinates will be written by template, so drop here to avoid possible conflicts
    output = output.drop_vars(["latitude", "longitude"])

    # MetView creates temporary grib files which need to be deleted manually
    for fs in [
        vort_div_fieldset,
        winds_spharm,
        fieldset,
        fieldset_gg,
        fieldset_2d,
        fieldset_2d_gg,
        winds_gg,
        thicknesses_fs,
    ]:
        _delete_fs(fs)

    return output


def process_native_data(
    key,
    ds,
    output_grid=DEFAULT_OUTPUT_GRID,
    output_layer_indices=DEFAULT_OUTPUT_LAYER_INDICES,
):
    output = _process_native_data(ds, output_grid, output_layer_indices)
    new_key = key.replace(
        offsets={"time": key.offsets["time"], "latitude": 0, "longitude": 0},
        vars=frozenset(output.keys()),
    )
    return new_key, output


def _split_and_average_over_forecast_hour(ds, **merge_kwargs):
    """Convert forecast_initial_time to time and average over forecast_hour."""
    xr.set_options(keep_attrs=True)
    logging.info("Splitting and averaging over forecast_hour")
    first = ds.isel(forecast_hour=slice(0, 6))
    second = ds.isel(forecast_hour=slice(6, 12))
    second["forecast_hour"] = first.forecast_hour
    # adjust times to label by end of the averaging period
    period = np.timedelta64(6, "h")
    first["forecast_initial_time"] = first.forecast_initial_time + period
    second["forecast_initial_time"] = second.forecast_initial_time + 2 * period

    first = first.rename({"forecast_initial_time": "time"})
    second = second.rename({"forecast_initial_time": "time"})
    output_ds = xr.merge([first, second], join="outer", **merge_kwargs)

    output_ds = output_ds.mean("forecast_hour")
    return output_ds


def split_and_average_over_forecast_hour(key, ds):
    output_ds = _split_and_average_over_forecast_hour(ds)
    # assuming we are splitting forecast_initial_time span by two. I.e. into a 6-hourly
    # time coordinate.
    new_key = key.with_offsets(
        time=2 * key.offsets["forecast_initial_time"],
        forecast_initial_time=None,
        forecast_hour=None,
    )
    return new_key, output_ds


def _regrid_quarter_degree(ds, output_grid):
    for name, attrs in DESIRED_ATTRS.items():
        if name in ds:
            ds[name] = ds[name].assign_attrs(**attrs)

    # metview chokes regridding length 1 time dimension data
    if ds.sizes.get("time", None) == 1:
        ds = ds.squeeze("time")
        restore_time = True
    else:
        restore_time = False

    # regrid to desired output grid
    regridded = xr.Dataset()
    for name in ds.data_vars:
        logging.info(f"Regridding {name} to output grid")
        fieldset = metview.dataset_to_fieldset(ds[[name]].load())
        fieldset_regridded = metview.regrid(data=fieldset, grid=output_grid)
        # for some reason, metview always sets the name to "t" when regridding
        # this may have something to do with the attrs of the input dataset
        regridded[name] = _to_dataarray(fieldset_regridded, "t")
        regridded[name].attrs = ds[name].attrs
        _delete_fs(fieldset)
        _delete_fs(fieldset_regridded)

    # time gets added back in for some reason
    regridded = regridded.drop_vars("time", errors="ignore")

    regridded = _adjust_latlon(regridded)

    # drop these scalar coords that get added by metview
    regridded = regridded.drop_vars(SCALAR_COORDS_TO_DROP, errors="ignore")

    if restore_time:
        regridded = regridded.expand_dims("time", axis=0)

    return regridded


def _adjust_latlon(ds):
    """Linearly interpolate to centerpoint between longitudes and flip latitude."""
    longitude_shift = 0.5 * (ds.longitude.values[1] - ds.longitude.values[0])
    # add cyclic point to avoid extrapolation
    cyclic_point = ds.isel(longitude=0)
    cyclic_point["longitude"] = 360 + cyclic_point.longitude
    ds = xr.concat([ds, cyclic_point], dim="longitude")
    output = ds.rolling(dim={"longitude": 2}).mean()
    # outputs of rolling mean are labeled by right side of window so first value is NaN
    output = output.isel(longitude=slice(1, None))
    output["longitude"] = output.longitude - longitude_shift
    output = output.reindex(latitude=output.latitude[::-1])
    return output


def _process_quarter_degree_data_mean_flux(ds, output_grid):
    logging.info("Processing 'mean-flux' quarter degree data")
    xr.set_options(keep_attrs=True)
    output = xr.Dataset()
    output["DSWRFtoa"] = ds.MTDWSWRF
    output["USWRFtoa"] = ds.MTDWSWRF - ds.MTNSWRF
    output["ULWRFtoa"] = -ds.MTNLWRF
    output["DSWRFsfc"] = ds.MSDWSWRF
    output["USWRFsfc"] = ds.MSDWSWRF - ds.MSNSWRF
    output["DLWRFsfc"] = ds.MSDWLWRF
    output["ULWRFsfc"] = ds.MSDWLWRF - ds.MSNLWRF
    output["SHTFLsfc"] = -ds.MSSHF  # opposite sign convention as FV3
    output["LHTFLsfc"] = -ds.MSLHF  # opposite sign convention as FV3
    output["PRATEsfc"] = ds.MTPR
    output["tendency_of_total_water_path_due_to_advection"] = -ds.MVIMD
    regridded = _regrid_quarter_degree(output, output_grid)

    # coordinates will be written by template, so drop here to avoid possible conflicts
    regridded = regridded.drop_vars(["latitude", "longitude"])

    return regridded


def _process_quarter_degree_data_sfc_an(ds, invariant_ds, output_grid):
    logging.info("Processing 'surface analysis' quarter degree data")
    xr.set_options(keep_attrs=True)
    output = xr.Dataset()
    output["sea_ice_fraction"] = ds.CI.fillna(0.0)
    output["soil_moisture_0"] = ds.SWVL1
    output["soil_moisture_1"] = ds.SWVL2
    output["soil_moisture_2"] = ds.SWVL3
    output["soil_moisture_3"] = ds.SWVL4
    regridded = _regrid_quarter_degree(output, output_grid)

    # coordinates will be written by template, so drop here to avoid possible conflicts
    regridded = regridded.drop_vars(["latitude", "longitude"])
    invariant_ds = invariant_ds.drop_vars(["latitude", "longitude"])

    regridded["ocean_fraction"] = (
        1 - invariant_ds.land_fraction - regridded.sea_ice_fraction
    )

    # In regridded data, sometimes land_fraction + sea_ice_fraction is greater than 1 so
    # we make a correction here to ensure there are no negative ocean_fraction values.
    # In principle, could result in negative sea_ice_fraction values, but we don't
    # find this to happen in practice.
    negative_ocean = xr.where(regridded.ocean_fraction < 0, regridded.ocean_fraction, 0)
    regridded["ocean_fraction"] -= negative_ocean
    regridded["sea_ice_fraction"] += negative_ocean

    for name in ["ocean_fraction", "sea_ice_fraction"]:
        regridded[name] = regridded[name].assign_attrs(DESIRED_ATTRS[name])

    return regridded


def _process_quarter_degree_data_invariant(ds, output_grid):
    logging.info("Renaming and fixing sign conventions for lat-lon invariant data")

    output = xr.Dataset()
    output["HGTsfc"] = ds.Z / GRAVITY
    output["land_fraction"] = ds.LSM

    regridded = _regrid_quarter_degree(output, output_grid)

    return regridded


def process_quarter_degree_data_mean_flux(key, ds, output_grid=DEFAULT_OUTPUT_GRID):
    output_ds = _process_quarter_degree_data_mean_flux(ds, output_grid)
    new_key = key.replace(vars=frozenset(output_ds.keys()))
    return new_key, output_ds


def process_quarter_degree_data_sfc_an(
    key, ds, invariant_ds=None, output_grid=DEFAULT_OUTPUT_GRID
):
    output_ds = _process_quarter_degree_data_sfc_an(ds, invariant_ds, output_grid)
    new_key = key.replace(vars=frozenset(output_ds.keys()))
    return new_key, output_ds


def _get_vertical_coordinate(
    ds: xr.Dataset, name: str, output_layer_indices: Sequence[int]
) -> xr.Dataset:
    """Get the ak/bk vertical coordinate on coarse layer interfaces.

    Assuming that ds[name] is a 3D variable which includes
    the vertical coordinate as an attribute named 'GRIB_pv'.
    """
    hybrid_sigma_values = ds[name].attrs["GRIB_pv"]
    ak = hybrid_sigma_values[: N_INPUT_LAYERS + 1]
    bk = hybrid_sigma_values[N_INPUT_LAYERS + 1 :]
    ak_coarse = [ak[i] for i in output_layer_indices]
    bk_coarse = [bk[i] for i in output_layer_indices]
    ak_coarse_ds = xr.Dataset({f"ak_{i}": value for i, value in enumerate(ak_coarse)})
    bk_coarse_ds = xr.Dataset({f"bk_{i}": value for i, value in enumerate(bk_coarse)})
    for name in ak_coarse_ds:
        ak_coarse_ds[name].attrs = {"long_name": f"ak", "units": "Pa"}
    for name in bk_coarse_ds:
        bk_coarse_ds[name].attrs = {"long_name": f"bk", "units": ""}
    return xr.merge([ak_coarse_ds, bk_coarse_ds])


def _make_template(
    ds_native,
    ds_meanflux,
    ds_quarter_degree_sfc,
    ds_quarter_degree_invariant,
    ds_google_latlon,
    ds_co2,
    ds_akbk,
    output_chunks,
    reuse_template,
    output_grid,
    output_layer_indices,
):
    """Here we (mostly) lazily process the data to make a reference zarr store
    for the output. This function mirrors what the pipeline does."""

    akbk_names = list(ds_akbk.keys())
    if reuse_template:
        logging.info(f"Using existing template at {TEMPLATE_PATH}")
        ds_regridded = xr.open_zarr(TEMPLATE_PATH).load()
        inv_fields = ds_regridded[["land_fraction", "HGTsfc"] + akbk_names]
    else:
        logging.info(f"Creating new template at {TEMPLATE_PATH}")
        ds_reshaped = _split_and_average_over_forecast_hour(
            ds_meanflux.isel(forecast_initial_time=0), compat="override"
        )
        ds_mean_flux_regridded = _process_quarter_degree_data_mean_flux(
            ds_reshaped, output_grid
        )
        ds_invariant_regridded = _process_quarter_degree_data_invariant(
            ds_quarter_degree_invariant, output_grid
        )
        ds_sfc_an_regridded = _process_quarter_degree_data_sfc_an(
            ds_quarter_degree_sfc.isel(time=0), ds_invariant_regridded, output_grid
        )
        ds_native_regridded = _process_native_data(
            ds_native.isel(time=0), output_grid, output_layer_indices
        )
        ds_google_latlon_regridded = _process_pressure_level_data(
            ds_google_latlon.isel(time=0), output_grid
        )
        ds_regridded = xr.merge(
            [
                ds_mean_flux_regridded,
                ds_sfc_an_regridded,
                ds_native_regridded,
                ds_invariant_regridded,
                ds_google_latlon_regridded,
                ds_akbk,
            ],
            compat="override",
            join="outer",
        ).squeeze()

        ds_regridded = ds_regridded.chunk({"latitude": -1, "longitude": -1})
        ds_regridded.to_zarr(TEMPLATE_PATH, mode="w")
        inv_fields = xr.merge([ds_invariant_regridded, ds_akbk])

    # manually expand time dim to include full time coordinate
    desired_time = ds_native.time.drop_vars(SCALAR_COORDS_TO_DROP, errors="ignore")
    template = xbeam.make_template(ds_regridded.drop_vars("time"))
    template = template.expand_dims(dim={"time": desired_time}, axis=0)
    template = template.chunk(output_chunks)

    inv_fields = inv_fields.drop_vars("time", errors="ignore")
    # includes regridded invariant field because ocean_fraction depends on
    # land fraction and temporally variable sea ice fraction
    # this will get written eagerly since it is not chunked
    template = template.update(inv_fields)
    template = template.update(ds_co2)

    return template, inv_fields


def _get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "output_path", type=str, help="Output path for the processed zarr dataset"
    )
    parser.add_argument("start_time", type=str, help="Desired start of output dataset")
    parser.add_argument("end_time", type=str, help="Desired end of output dataset")
    parser.add_argument(
        "--output_grid",
        type=str,
        default="F90",
        help=(
            "Output grid specification according to ECMWF nomenclature. E.g. 'F90' for "
            f"1° Gaussian Grid. See more information at {GRID_DOCS_URL}"
        ),
    )
    parser.add_argument(
        "--output_time_chunksize",
        type=int,
        default=50,
        help="Number of times per output chunk.",
    )
    parser.add_argument(
        "--ncar_process_time_chunksize",
        type=int,
        default=10,
        help=(
            "Time chunk size for intermediate regridding step for all data from "
            "NCAR-sourced datasets (mean flux / surface analysis / invariant). Must "
            "be a multiple of 2 and divide evenly into output_time_chunksize."
        ),
    )
    parser.add_argument(
        "--reuse_template",
        action="store_true",
        help=(
            "Reuse the existing template at {TEMPLATE_PATH}. This can be helpful to "
            "speed up debugging but should not be used for production runs."
        ),
    )
    parser.add_argument(
        "--output-layer-indices",
        type=int,
        nargs="+",
        default=DEFAULT_OUTPUT_LAYER_INDICES,
        help=(
            "Specify the ERA5 layer indices to use when defining the vertically "
            "coarsened output levels."
        ),
    )
    return parser


def main():
    parser = _get_parser()
    args, pipeline_args = parser.parse_known_args()

    # desired start/end of output dataset, inclusive
    start_time = datetime.datetime.strptime(args.start_time, "%Y-%m-%dT%H:%M:%S")
    end_time = datetime.datetime.strptime(args.end_time, "%Y-%m-%dT%H:%M:%S")

    # to properly align mean flux with the output time, these requirements are necessary
    assert start_time.hour == 0 or start_time.hour == 12
    assert end_time.hour == 6 or end_time.hour == 18

    # mean flux forecast initial times are at 6Z and 18Z, and we label by the
    # end of the time-averaging periods
    start_time_mean_flux = start_time - datetime.timedelta(hours=6)
    end_time_mean_flux = end_time - datetime.timedelta(hours=12)

    regular_time_slice = {"time": slice(start_time, end_time, TIME_STEP)}
    forecast_time_slice = {
        "forecast_initial_time": slice(start_time_mean_flux, end_time_mean_flux)
    }
    sel_indices = {
        WIND_TEMP: regular_time_slice,
        SURFACE: regular_time_slice,
        SURFACE_REANALYSIS: regular_time_slice,
        MOISTURE: regular_time_slice,
        INVARIANT: dict(time="1979-01-01T00:00:00"),
        SURFACE_ANALYSIS_LATLON: regular_time_slice,
        MEAN_FLUX: forecast_time_slice,
        GOOGLE_LATLON: regular_time_slice,
    }

    msg = (
        "ncar_process_time_chunksize must be a multiple of 2, "
        f"got {args.ncar_process_time_chunksize}"
    )
    assert args.ncar_process_time_chunksize % 2 == 0, msg
    msg = (
        "output_time_chunksize must be a multiple of ncar_process_time_chunksize, "
        f"got {args.output_time_chunksize} and {args.ncar_process_time_chunksize}"
    )
    assert args.output_time_chunksize % args.ncar_process_time_chunksize == 0, msg
    output_chunks = {"time": args.output_time_chunksize}
    ncar_process_chunks = {"time": args.ncar_process_time_chunksize}

    logging.info("Opening datasets")
    ds_native = open_native_datasets(sel_indices)
    ds_meanflux = open_meanflux_dataset(sel_indices)
    ds_quarter_degree_sfc, ds_quarter_degree_inv = open_quarter_degree_datasets(
        sel_indices
    )
    ds_google_latlon = open_google_latlon_dataset(sel_indices)
    ds_co2 = open_co2_dataset(start_time, end_time)
    logging.info("Getting vertical coordinate")
    ds_akbk = _get_vertical_coordinate(ds_native, "t", args.output_layer_indices)

    logging.info("Generating template")
    template, ds_pt25deg_inv_regridded = _make_template(
        ds_native,
        ds_meanflux,
        ds_quarter_degree_sfc,
        ds_quarter_degree_inv,
        ds_google_latlon,
        ds_co2,
        ds_akbk,
        output_chunks,
        args.reuse_template,
        args.output_grid,
        args.output_layer_indices,
    )

    logging.info("Template finished generating. Starting pipeline.")
    with beam.Pipeline(options=PipelineOptions(pipeline_args)) as p:
        (
            p
            | xbeam.DatasetToChunks(ds_meanflux, chunks={"forecast_initial_time": 1})
            | beam.MapTuple(split_and_average_over_forecast_hour)
            | xbeam.ConsolidateChunks(ncar_process_chunks)
            | beam.MapTuple(
                process_quarter_degree_data_mean_flux, output_grid=args.output_grid
            )
            | "mean_flux_ConsolidateChunks" >> xbeam.ConsolidateChunks(output_chunks)
            | "mean_flux_to_zarr"
            >> xbeam.ChunksToZarr(args.output_path, template, output_chunks)
        )

        (
            p
            | "qd_DatasetToChunks"
            >> xbeam.DatasetToChunks(ds_quarter_degree_sfc, chunks=ncar_process_chunks)
            | beam.MapTuple(
                process_quarter_degree_data_sfc_an,
                invariant_ds=ds_pt25deg_inv_regridded,
                output_grid=args.output_grid,
            )
            | "qd_ConsolidateChunks" >> xbeam.ConsolidateChunks(output_chunks)
            | "qd_to_zarr"
            >> xbeam.ChunksToZarr(args.output_path, template, output_chunks)
        )

        (
            p
            | "pl_DatasetToChunks"
            >> xbeam.DatasetToChunks(ds_google_latlon, chunks=ncar_process_chunks)
            | beam.MapTuple(process_pressure_level_data)
            | "pl_ConsolidateChunks" >> xbeam.ConsolidateChunks(output_chunks)
            | "pl_to_zarr"
            >> xbeam.ChunksToZarr(args.output_path, template, output_chunks)
        )

        (
            p
            | "native_DatasetToChunks"
            >> xbeam.DatasetToChunks(ds_native, chunks={"time": 1})
            | beam.MapTuple(
                process_native_data,
                output_grid=args.output_grid,
                output_layer_indices=args.output_layer_indices,
            )
            | "native_ConsolidateChunks" >> xbeam.ConsolidateChunks(output_chunks)
            | "native_to_zarr"
            >> xbeam.ChunksToZarr(args.output_path, template, output_chunks)
        )


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    logging.getLogger().setLevel(logging.INFO)
    main()



import argparse
import logging

import dacite
import dacite.exceptions
import yaml

from fme.ace.inference.evaluator import InferenceEvaluatorConfig
from fme.ace.inference.inference import InferenceConfig
from fme.ace.stepper import SingleModuleStepperConfig
from fme.ace.train.train_config import TrainConfig
from fme.core.config import update_dict_with_dotlist

CONFIG_CHOICES = ["train", "inference", "evaluator"]

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "path", type=str, help="Path to the train or inference config file."
    )
    parser.add_argument(
        "--inference",
        action="store_true",
        help=(
            "Deprecated, use --config_type evaluator to validate an evaluator config."
        ),
    )
    parser.add_argument(
        "--config_type",
        type=str,
        choices=CONFIG_CHOICES,
        default="train",
        help=("Indicates the kind of configuration being validated."),
    )
    parser.add_argument(
        "--override",
        nargs="*",
        help="A dotlist of key=value pairs to override the config. "
        "For example, --override a.b=1 c=2, where a dot indicates nesting.",
    )
    args = parser.parse_args()

    if args.inference:
        logging.warning(
            "The --inference flag is deprecated. "
            "Use --config_type evaluator to validate an evaluator config."
        )
        config_type = "evaluator"
    else:
        config_type = args.config_type

    with open(args.path) as f:
        config_data = yaml.load(f, Loader=yaml.CLoader)

    config_data = update_dict_with_dotlist(config_data, args.override)
    if config_type == "evaluator":
        dacite.from_dict(
            data_class=InferenceEvaluatorConfig,
            data=config_data,
            config=dacite.Config(strict=True),
        )
    elif config_type == "inference":
        dacite.from_dict(
            data_class=InferenceConfig,
            data=config_data,
            config=dacite.Config(strict=True),
        )
    elif config_type == "train":
        try:
            dacite.from_dict(
                data_class=TrainConfig,
                data=config_data,
                config=dacite.Config(strict=True),
            )
        except dacite.exceptions.UnionMatchError as err:
            if "checkpoint_path" not in config_data["stepper"]:
                dacite.from_dict(
                    data_class=SingleModuleStepperConfig,
                    data=config_data["stepper"],
                    config=dacite.Config(strict=True),
                )
            # if there was no issue for SingleModuleStepperConfig, raise original error
            raise err
    else:
        raise ValueError(
            f"Invalid config type: {config_type}, expected one of {CONFIG_CHOICES}"
        )



import dataclasses
from collections.abc import Iterable

import dacite
import pytest
import torch

from .module import ModuleConfig, ModuleSelector


class MockModule(torch.nn.Module):
    def __init__(self, param_shapes: Iterable[tuple[int, ...]]):
        super().__init__()
        for i, shape in enumerate(param_shapes):
            setattr(self, f"param{i}", torch.nn.Parameter(torch.randn(shape)))


@ModuleSelector.register("mock")
@dataclasses.dataclass
class MockModuleBuilder(ModuleConfig):
    param_shapes: list[tuple[int, ...]]

    def build(self, n_in_channels, n_out_channels, img_shape):
        return MockModule(self.param_shapes)

    @classmethod
    def from_state(cls, state):
        return cls(state["param_shapes"])

    def get_state(self):
        return {
            "param_shapes": self.param_shapes,
        }


def test_register():
    """Make sure that the registry is working as expected."""
    selector = ModuleSelector(type="mock", config={"param_shapes": [(1, 2, 3)]})
    module = selector.build(1, 1, (16, 32))
    assert isinstance(module, MockModule)


def test_module_selector_raises_with_bad_config():
    with pytest.raises(dacite.UnexpectedDataError):
        ModuleSelector(type="mock", config={"non_existent_key": 1})



import abc
from collections.abc import Mapping

import torch
import xarray as xr

from fme.ace.aggregator.plotting import plot_paneled_data
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.distributed import Distributed
from fme.core.ensemble import get_crps
from fme.core.gridded_ops import GriddedOperations
from fme.core.typing_ import EnsembleTensorDict, TensorMapping


def get_gen_shape(gen_data: TensorMapping):
    for name in gen_data:
        return gen_data[name].shape


def get_one_step_ensemble_aggregator(
    gridded_operations: GriddedOperations,
    target_time: int = 1,
    log_mean_maps: bool = True,
    metadata: Mapping[str, VariableMetadata] | None = None,
) -> "SelectStepEnsembleAggregator":
    return SelectStepEnsembleAggregator(
        aggregator=_EnsembleAggregator(
            gridded_operations=gridded_operations,
            log_mean_maps=log_mean_maps,
            metadata=metadata,
        ),
        i_target_time=target_time,
    )


class ReducedMetric(abc.ABC):
    """
    Used to record a metric value on batches of data (potentially out-of-memory)
    and then get the final metric at the end.
    """

    @abc.abstractmethod
    def record(self, target: torch.Tensor, gen: torch.Tensor):
        """
        Update metric for a batch of data.
        """
        ...

    @abc.abstractmethod
    def get(self) -> torch.Tensor:
        """
        Get the final metric value.
        """
        ...


class CRPSMetric(ReducedMetric):
    def __init__(self):
        self._total = None
        self._n_batches = 0

    def record(self, target: torch.Tensor, gen: torch.Tensor):
        crps = get_crps(gen=gen, target=target, alpha=0.95).mean(dim=(0, 1))
        if self._total is None:
            self._total = crps
        else:
            self._total += crps
        self._n_batches += 1

    def get(self) -> torch.Tensor:
        if self._total is None:
            raise ValueError("No batches have been recorded.")
        return self._total / self._n_batches


class SSRBiasMetric(ReducedMetric):
    """
    Computes the spread-skill ratio bias (equal to (stdev / rmse) - 1).
    """

    def __init__(self):
        self._total_mse = None
        self._total_variance = None
        self._n_batches = 0

    def record(self, target: torch.Tensor, gen: torch.Tensor):
        mse = ((gen - target) ** 2).mean(dim=(0, 1, 2))  # batch, ensemble, time
        variance = gen.var(dim=1, unbiased=True).mean(dim=(0, 1))
        self._add_mse(mse)
        self._add_variance(variance)
        self._n_batches += 1

    def _add_mse(self, mse: torch.Tensor):
        if self._total_mse is None:
            self._total_mse = torch.zeros_like(mse)
        self._total_mse += mse

    def _add_variance(self, variance: torch.Tensor):
        if self._total_variance is None:
            self._total_variance = torch.zeros_like(variance)
        self._total_variance += variance

    def get(self) -> torch.Tensor:
        if self._total_mse is None or self._total_variance is None:
            raise ValueError("No batches have been recorded.")
        spread = self._total_variance.sqrt()
        # must remove the component of the MSE that is due to the
        # variance of the generated values
        skill = (self._total_mse - self._total_variance).sqrt()
        return spread / skill - 1


class _EnsembleAggregator:
    """
    Aggregator for ensemble-based metrics.
    """

    def __init__(
        self,
        gridded_operations: GriddedOperations,
        log_mean_maps: bool = True,
        metadata: Mapping[str, VariableMetadata] | None = None,
    ):
        """
        Args:
            gridded_operations: Gridded operations to use.
            log_mean_maps: Whether to log mean maps.
            metadata: Mapping of variable names their metadata that will
                used in generating logged image captions.
        """
        self._gridded_operations = gridded_operations
        self._n_batches = 0
        self._variable_metrics: dict[str, dict[str, ReducedMetric]] | None = None
        self._dist = Distributed.get_instance()
        self._log_mean_maps = log_mean_maps
        self._metadata = metadata
        self._diverging_metrics = {"ssr_bias"}

    def _get_variable_metrics(self, gen_data: TensorMapping):
        if self._variable_metrics is None:
            self._variable_metrics = {
                "crps": {},
                "ssr_bias": {},
            }
            for key in gen_data:
                self._variable_metrics["crps"][key] = CRPSMetric()
                self._variable_metrics["ssr_bias"][key] = SSRBiasMetric()
        return self._variable_metrics

    @torch.no_grad()
    def record_batch(
        self,
        target_data: EnsembleTensorDict,
        gen_data: EnsembleTensorDict,
    ):
        """
        Record a batch of data.

        Args:
            target_data: Target data, of shape [batch, ensemble, time, ...].
            gen_data: Generated data, of shape [batch, ensemble, time, ...].
        """
        variable_metrics = self._get_variable_metrics(gen_data)
        for metric in variable_metrics:
            for name in gen_data:
                variable_metrics[metric][name].record(
                    target=target_data[name],
                    gen=gen_data[name],
                )
        self._n_batches += 1

    def _get_caption(self, name: str) -> str:
        if self._metadata is not None and name in self._metadata:
            caption_name = self._metadata[name].long_name
            units = self._metadata[name].units
        else:
            caption_name, units = name, "unknown_units"
        caption = f"{caption_name} ({units})"
        return caption

    def _get_data(self):
        if self._variable_metrics is None or self._n_batches == 0:
            raise ValueError("No batches have been recorded.")
        data: dict[str, torch.Tensor] = {}
        for metric in sorted(self._variable_metrics):
            for key in sorted(self._variable_metrics[metric]):
                metric_value = self._dist.reduce_mean(
                    self._variable_metrics[metric][key].get()
                )
                data[f"{metric}/{key}"] = (
                    self._gridded_operations.area_weighted_mean(metric_value, name=key)
                    .cpu()
                    .numpy()
                )
                if self._log_mean_maps:
                    data[f"{metric}/mean_map/{key}"] = plot_paneled_data(
                        [[metric_value.cpu().numpy()]],
                        diverging=key in self._diverging_metrics,
                        caption=self._get_caption(key),
                    )
        return data

    @torch.no_grad()
    def get_logs(self, label: str):
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
        """
        return {
            f"{label}/{key}": data for key, data in sorted(self._get_data().items())
        }

    @torch.no_grad()
    def get_dataset(self) -> xr.Dataset:
        data = self._get_data()
        data = {key.replace("/", "-"): data[key] for key in data}
        data_vars = {}
        for key, value in data.items():
            data_vars[key] = xr.DataArray(value)
        return xr.Dataset(data_vars=data_vars)


class SelectStepEnsembleAggregator:
    """
    Wraps an aggregator that takes a time dimension, and records
    metrics for a specific time step.
    """

    def __init__(
        self,
        aggregator: _EnsembleAggregator,
        i_target_time: int,
    ):
        """
        Args:
            aggregator: Aggregator to wrap.
            i_target_time: Global time index of the target time step.
        """
        self._aggregator = aggregator
        self._i_target_time = i_target_time

    def record_batch(
        self,
        target_data: EnsembleTensorDict,
        gen_data: EnsembleTensorDict,
        i_time_start: int = 0,
    ):
        """
        Record a specific global timestep of data.

        Call does nothing if the target time is not in this batch.

        Args:
            target_data: Target data, of shape [batch, ensemble, time, ...].
            gen_data: Generated data, of shape [batch, ensemble, time, ...].
            i_time_start: Global time index of the first time step in the batch.
        """
        n_timesteps = next(iter(target_data.values())).shape[2]
        if (
            self._i_target_time >= i_time_start
            and self._i_target_time < i_time_start + n_timesteps
        ):
            batch_i_target_time = self._i_target_time - i_time_start
            target_data = EnsembleTensorDict(
                {
                    key: value[:, :, batch_i_target_time : batch_i_target_time + 1, ...]
                    for key, value in target_data.items()
                }
            )
            gen_data = EnsembleTensorDict(
                {
                    key: value[:, :, batch_i_target_time : batch_i_target_time + 1, ...]
                    for key, value in gen_data.items()
                }
            )
            self._aggregator.record_batch(
                target_data,
                gen_data,
            )

    def get_logs(self, label: str):
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
        """
        return self._aggregator.get_logs(label)

    def get_dataset(self) -> xr.Dataset:
        return self._aggregator.get_dataset()



import pytest
import torch

from .stacker import Stacker, unstack

_PREFIX_MAP = {"a": ["aa_", "ab_"], "b": ["b_", "bb_"], "c": ["ca", "cb"]}


@pytest.mark.parametrize("prefix_map", [_PREFIX_MAP])
@pytest.mark.parametrize(
    "data_names, expected_level_names",
    [
        (
            ["aa_0", "aa_1", "b_00", "b_01", "b", "ca"],
            {"a": ["aa_0", "aa_1"], "b": ["b_00", "b_01"], "c": ["ca"]},
        ),
        (
            ["ab_1", "ab_0", "bb_1", "bb_2", "cb"],
            {"a": ["ab_0", "ab_1"], "b": (ValueError, "Missing level 0"), "c": ["cb"]},
        ),
        (
            ["ab_0", "b_", "cc"],
            {
                "a": ["ab_0"],
                "b": ["b_"],
                "c": (KeyError, "No prefix associated with 'c'"),
            },
        ),
    ],
)
def test_get_all_level_names(prefix_map, data_names, expected_level_names):
    data = {name: torch.zeros(2, 2) for name in data_names}
    stacker = Stacker(prefix_map)
    for standard_name in expected_level_names.keys():
        expected_names = expected_level_names[standard_name]
        if isinstance(expected_names, tuple):
            expected_error, message = expected_names
            with pytest.raises(expected_error, match=message):
                stacker.get_all_level_names(standard_name, data)
        else:
            assert (
                stacker.get_all_level_names(standard_name, data)
                == expected_level_names[standard_name]
            )


@pytest.mark.parametrize("prefix_map", [_PREFIX_MAP])
@pytest.mark.parametrize(
    "data_names, expected_level_names",
    [
        (
            ["aa_0", "aa_1", "b_00", "b_01", "b", "ca"],
            {"a": ["aa_0", "aa_1"], "b": ["b_00", "b_01"], "c": ["ca"]},
        ),
        (
            ["ab_0", "ab_1", "bb_2", "bb_1", "cb"],
            {"a": ["ab_0", "ab_1"], "c": ["cb"]},
        ),
        (
            ["ab_0", "b_", "cc"],
            {"a": ["ab_0"], "b": ["b_"]},
        ),
    ],
)
def test_stack_unstack(prefix_map, data_names, expected_level_names):
    torch.manual_seed(0)
    data = {name: torch.rand(2, 2) for name in data_names}
    stacker = Stacker(prefix_map)
    for standard_name in expected_level_names.keys():
        level_names = expected_level_names[standard_name]
        if len(level_names) == 1:
            expected_stacked = data[level_names[0]].unsqueeze(-1)
        else:
            expected_stacked = torch.stack([data[name] for name in level_names], dim=-1)
        stacked = stacker(standard_name, data)
        assert torch.allclose(stacked, expected_stacked)
        unstacked = unstack(
            stacked, names=stacker.get_all_level_names(standard_name, data)
        )
        for name in level_names:
            assert name in unstacked
            assert torch.allclose(unstacked[name], data[name])


@pytest.mark.parametrize(
    "data_names, expected_prefix_map, expected_level_names",
    [
        (
            ["aa_1", "aa_0", "b_00", "b_01", "b", "ca"],
            {"aa_": ["aa_"], "b_": ["b_"], "b": ["b"], "ca": ["ca"]},
            {"aa_": ["aa_0", "aa_1"], "b_": ["b_00", "b_01"], "b": ["b"], "ca": ["ca"]},
        ),
        (
            ["ab_0", "ab_1", "bb_1", "bb_2", "cb"],
            {"ab_": ["ab_"], "bb_": ["bb_"], "cb": ["cb"]},
            {
                "ab_": ["ab_0", "ab_1"],
                "bb_": (ValueError, "Missing level 0"),
                "cb": ["cb"],
            },
        ),
        (
            ["ab_0", "b_", "cc"],
            {"ab_": ["ab_"], "b_": ["b_"], "cc": ["cc"]},
            {"ab_": ["ab_0"], "b_": ["b_"], "cc": ["cc"]},
        ),
    ],
)
def test_inferred_stacker(data_names, expected_prefix_map, expected_level_names):
    data = {name: torch.rand(2, 2) for name in data_names}
    stacker = Stacker()
    with pytest.raises(RuntimeError, match="hasn't yet been built"):
        stacker("a", data)
    stacker.infer_prefix_map(data.keys())
    with pytest.raises(RuntimeError, match="has already been built"):
        stacker.infer_prefix_map(data.keys())
    assert stacker.prefix_map == expected_prefix_map
    standard_names = stacker.standard_names
    assert standard_names == list(expected_prefix_map.keys())
    for standard_name in stacker.standard_names:
        expected_levels = expected_level_names[standard_name]
        if isinstance(expected_levels, tuple):
            expected_error, message = expected_levels
            with pytest.raises(expected_error, match=message):
                _ = stacker(standard_name, data)
        else:
            stacked = stacker(standard_name, data)
            tensors = [data[name] for name in expected_levels]
            expected_stacked = torch.stack(tensors, dim=-1)
            assert torch.allclose(stacked, expected_stacked)



import contextlib
import dataclasses
import logging
import os
from collections.abc import Mapping
from typing import Any

from fme.core.distributed import Distributed
from fme.core.wandb import WandB

ENV_VAR_NAMES = (
    "BEAKER_EXPERIMENT_ID",
    "SLURM_JOB_ID",
    "SLURM_JOB_USER",
    "FME_TRAIN_DIR",
    "FME_VALID_DIR",
    "FME_STATS_DIR",
    "FME_CHECKPOINT_DIR",
    "FME_OUTPUT_DIR",
    "FME_IMAGE",
)

DEFAULT_TMP_DIR = "/tmp"


@dataclasses.dataclass
class LoggingConfig:
    """
    Configuration for logging.

    Parameters:
        project: Name of the project in Weights & Biases.
        entity: Name of the entity in Weights & Biases.
        log_to_screen: Whether to log to the screen.
        log_to_file: Whether to log to a file.
        log_to_wandb: Whether to log to Weights & Biases.
        log_format: Format of the log messages.
        level: Sets the logging level.
    """

    project: str = "ace"
    entity: str = "ai2cm"
    log_to_screen: bool = True
    log_to_file: bool = True
    log_to_wandb: bool = True
    log_format: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    level: str | int = logging.INFO

    def __post_init__(self):
        self._dist = Distributed.get_instance()

    def configure_logging(self, experiment_dir: str, log_filename: str):
        """
        Configure the global `logging` module based on this LoggingConfig.
        """
        if self.log_to_screen and self._dist.is_root():
            logging.basicConfig(format=self.log_format, level=self.level)
        elif self._dist.is_root():
            logging.basicConfig(level=logging.WARNING)
        else:  # we are not root
            logging.basicConfig(level=logging.ERROR)
        logger = logging.getLogger()
        if self.log_to_file and self._dist.is_root():
            if not os.path.exists(experiment_dir):
                raise ValueError(
                    f"experiment directory {experiment_dir} does not exist, "
                    "cannot log files to it"
                )
            log_path = os.path.join(experiment_dir, log_filename)
            fh = logging.FileHandler(log_path)
            fh.setLevel(self.level)
            fh.setFormatter(logging.Formatter(self.log_format))
            logger.addHandler(fh)

    def configure_wandb(
        self,
        config: Mapping[str, Any],
        env_vars: Mapping[str, Any] | None = None,
        wandb_dir: str | None = DEFAULT_TMP_DIR,
        resumable: bool = True,
        resume: Any = None,
        **kwargs,
    ):
        if resume is not None:
            raise ValueError(
                "The 'resume' argument is no longer supported, "
                "please pass 'resumable' instead."
            )
        config_copy = {**config}
        if "environment" in config_copy:
            logging.warning(
                "Not recording environmental variables since 'environment' key is "
                "already present in config."
            )
        elif env_vars is not None:
            config_copy["environment"] = env_vars
        # must ensure wandb.configure is called before wandb.init
        wandb = WandB.get_instance()
        wandb.configure(log_to_wandb=self.log_to_wandb)
        wandb.init(
            config=config_copy,
            project=self.project,
            entity=self.entity,
            experiment_dir=config["experiment_dir"],
            resumable=resumable,
            dir=wandb_dir,
            **kwargs,
        )


def log_versions():
    import torch

    logging.info("--------------- Versions ---------------")
    logging.info("Torch: " + str(torch.__version__))
    logging.info("----------------------------------------")


def retrieve_env_vars(names=ENV_VAR_NAMES) -> dict[str, str]:
    """Return a dictionary of specific environmental variables."""
    output = {}
    for name in names:
        try:
            value = os.environ[name]
        except KeyError:
            logging.warning(f"Environmental variable {name} not found.")
        else:
            output[name] = value
            logging.info(f"Environmental variable {name}={value}.")
    return output


def log_beaker_url(beaker_id=None):
    """Log the Beaker ID and URL for the current experiment.

    beaker_id: The Beaker ID of the experiment. If None, uses the env variable
    `BEAKER_EXPERIMENT_ID`.

    Returns the Beaker URL.
    """
    if beaker_id is None:
        try:
            beaker_id = os.environ["BEAKER_EXPERIMENT_ID"]
        except KeyError:
            logging.warning("Beaker Experiment ID not found.")
            return None

    beaker_url = f"https://beaker.org/ex/{beaker_id}"
    logging.info(f"Beaker ID: {beaker_id}")
    logging.info(f"Beaker URL: {beaker_url}")
    return beaker_url


@contextlib.contextmanager
def log_level(level):
    """Temporarily set the log level of the global logger."""
    logger = logging.getLogger()  # presently, data loading uses the root logger
    old_level = logger.getEffectiveLevel()
    try:
        logger.setLevel(level)
        yield
    finally:
        logger.setLevel(old_level)



import numpy as np
import torch

from fme.ace.aggregator.inference.time_mean import (
    TimeMeanAggregator,
    TimeMeanEvaluatorAggregator,
)
from fme.core.device import get_device
from fme.core.gridded_ops import LatLonOperations


def test_rmse_of_time_mean_all_channels():
    torch.manual_seed(0)
    area_weights = torch.ones(1).to(get_device())
    agg = TimeMeanEvaluatorAggregator(
        LatLonOperations(area_weights),
        horizontal_dims=["lat", "lon"],
        target="norm",
    )
    target_data_norm = {
        "a": torch.ones([2, 3, 4, 4], device=get_device()),
        "b": torch.ones([2, 3, 4, 4], device=get_device()) * 3,
    }
    gen_data_norm = {
        "a": torch.ones([2, 3, 4, 4], device=get_device()) * 2.0,
        "b": torch.ones([2, 3, 4, 4], device=get_device()) * 5,
    }
    agg.record_batch(
        target_data=target_data_norm,
        gen_data=gen_data_norm,
        target_data_norm=target_data_norm,
        gen_data_norm=gen_data_norm,
    )
    logs = agg.get_logs(label="time_mean_norm")
    assert logs["time_mean_norm/rmse/a"] == 1.0
    assert logs["time_mean_norm/rmse/b"] == 2.0
    assert logs["time_mean_norm/rmse/channel_mean"] == 1.5


def test_custom_channel_mean_names():
    torch.manual_seed(0)
    area_weights = torch.ones(1).to(get_device())
    agg = TimeMeanEvaluatorAggregator(
        LatLonOperations(area_weights),
        horizontal_dims=["lat", "lon"],
        target="norm",
        channel_mean_names=["a"],
    )
    target_data_norm = {
        "a": torch.ones([2, 3, 4, 4], device=get_device()),
        "b": torch.ones([2, 3, 4, 4], device=get_device()) * 3,
    }
    gen_data_norm = {
        "a": torch.ones([2, 3, 4, 4], device=get_device()) * 2.0,
        "b": torch.ones([2, 3, 4, 4], device=get_device()) * 5,
    }
    agg.record_batch(
        target_data=target_data_norm,
        gen_data=gen_data_norm,
        target_data_norm=target_data_norm,
        gen_data_norm=gen_data_norm,
    )
    logs = agg.get_logs(label="time_mean_norm")
    assert logs["time_mean_norm/rmse/a"] == 1.0
    assert logs["time_mean_norm/rmse/b"] == 2.0
    assert logs["time_mean_norm/rmse/channel_mean"] == 1.0


def test_mean_all_channels_not_in_denorm():
    area_weights = torch.ones(1).to(get_device())
    agg = TimeMeanEvaluatorAggregator(
        LatLonOperations(area_weights),
        horizontal_dims=["lat", "lon"],
        target="denorm",
    )
    target_data = {
        "a": torch.ones([2, 3, 4, 4], device=get_device()),
        "b": torch.ones([2, 3, 4, 4], device=get_device()) * 3,
    }
    gen_data = {
        "a": torch.ones([2, 3, 4, 4], device=get_device()) * 2.0,
        "b": torch.ones([2, 3, 4, 4], device=get_device()) * 5,
    }
    agg.record_batch(
        target_data=target_data,
        gen_data=gen_data,
        target_data_norm=target_data,
        gen_data_norm=gen_data,
    )
    logs = agg.get_logs(label="time_mean")
    assert "time_mean/rmse/channel_mean" not in list(logs.keys())
    ds = agg.get_dataset()
    assert "bias_map-a" in ds
    assert np.all(ds["bias_map-a"].values == 1.0)


def test_bias_values():
    area_weights = torch.ones(1).to(get_device())
    agg = TimeMeanEvaluatorAggregator(
        LatLonOperations(area_weights),
        horizontal_dims=["lat", "lon"],
        target="denorm",
    )
    # use constant values so area-weighting doesn't matter
    target_data = {
        "a": (torch.rand(1) * torch.ones(size=[2, 3, 4, 5])).to(device=get_device()),
    }
    gen_data = {
        "a": (torch.rand(1) * torch.ones(size=[2, 3, 4, 5])).to(device=get_device()),
    }
    agg.record_batch(
        target_data=target_data,
        gen_data=gen_data,
        target_data_norm=target_data,
        gen_data_norm=gen_data,
    )
    ds = agg.get_dataset()
    assert "bias_map-a" in ds
    np.testing.assert_array_equal(
        ds["bias_map-a"].values,
        (
            gen_data["a"].cpu().numpy().mean(axis=(0, 1))
            - target_data["a"].cpu().numpy().mean(axis=(0, 1))
        ),
    )
    assert "gen_map-a" in ds
    np.testing.assert_array_equal(
        ds["gen_map-a"].values,
        (gen_data["a"].cpu().numpy().mean(axis=(0, 1))),
    )


def test_aggregator_mean_values():
    area_weights = torch.ones(1).to(get_device())
    agg = TimeMeanAggregator(LatLonOperations(area_weights))
    # use constant values so area-weighting doesn't matter
    data = {
        "a": (torch.rand(1) * torch.ones(size=[2, 3, 4, 5])).to(device=get_device()),
    }
    agg.record_batch(
        data=data,
        i_time_start=0,
    )
    ds = agg.get_dataset()
    assert "gen_map-a" in ds
    np.testing.assert_array_equal(
        ds["gen_map-a"].values,
        (data["a"].cpu().numpy().mean(axis=(0, 1))),
    )



import dataclasses
import datetime
from collections.abc import Callable, Mapping
from functools import partial
from typing import Any, Optional

import numpy as np
import torch
import xarray as xr
from matplotlib.figure import Figure

from fme.core.dataset.data_typing import VariableMetadata
from fme.core.device import get_device
from fme.core.distributed import Distributed
from fme.core.gridded_ops import GriddedOperations
from fme.core.typing_ import TensorMapping

from ..plotting import plot_mean_and_samples


class PairedGlobalMeanAnnualAggregator:
    def __init__(
        self,
        ops: GriddedOperations,
        timestep: datetime.timedelta,
        variable_metadata: Mapping[str, VariableMetadata] | None = None,
        monthly_reference_data: xr.Dataset | None = None,
    ):
        self._area_weighted_mean = ops.area_weighted_mean
        self.timestep = timestep
        self.variable_metadata = variable_metadata or {}
        self._target_aggregator = GlobalMeanAnnualAggregator(
            ops, timestep, variable_metadata
        )
        self._gen_aggregator = GlobalMeanAnnualAggregator(
            ops, timestep, variable_metadata
        )
        self._monthly_reference_data = monthly_reference_data
        self._variable_reference_data: dict[str, VariableReferenceData] = {}

    def _get_reference(self, name: str) -> Optional["VariableReferenceData"]:
        if self._monthly_reference_data is None:
            return None
        if name not in self._variable_reference_data:
            if name not in self._monthly_reference_data:
                return None
            area_weighted_mean = partial(self._area_weighted_mean, name=name)
            self._variable_reference_data[name] = process_monthly_reference(
                self._monthly_reference_data, area_weighted_mean, name
            )
        return self._variable_reference_data[name]

    @torch.no_grad()
    def record_batch(
        self,
        time: xr.DataArray,
        target_data: TensorMapping,
        gen_data: TensorMapping,
    ):
        """Record a batch of data for computing time variability statistics."""
        self._target_aggregator.record_batch(time, target_data)
        self._gen_aggregator.record_batch(time, gen_data)

    def _get_gathered_means(self) -> tuple[xr.Dataset, xr.Dataset] | None:
        """
        Gather the mean target and generated data across all processes.

        Returns:
            A tuple of the target and generated datasets, or None if this is not the
            root rank.
        """
        target = self._target_aggregator.get_gathered_means()
        gen = self._gen_aggregator.get_gathered_means()
        if target is None or gen is None:
            return None
        return target, gen

    @torch.no_grad()
    def get_logs(self, label: str) -> dict[str, Any]:
        gathered = self._get_gathered_means()
        if gathered is None:  # not the root rank
            return {}
        target, gen = gathered
        plots = {}
        metrics = {}

        for name in gen.data_vars.keys():
            if name == "counts":
                continue

            if name in self.variable_metadata:
                long_name = self.variable_metadata[name].long_name
                units = self.variable_metadata[name].units
            else:
                long_name = name
                units = "unknown units"

            fig = Figure()  # create directly for cleanup when it leaves scope
            ax = fig.add_subplot(1, 1, 1)  # Add an axes to the figure
            ref = self._get_reference(name)
            if ref is not None:
                if ref.mean.sizes["year"] > 1:
                    # dataarray.plot() does not work for singleton dimensions
                    ref.mean.plot(ax=ax, x="year", label="ref_mean", color="black")
                    ref.min.plot(
                        ax=ax, x="year", label="ref_min", color="grey", linestyle="--"
                    )
                    ref.max.plot(
                        ax=ax, x="year", label="ref_max", color="grey", linestyle="--"
                    )
            if gen.sizes["year"] > 1:
                target_ensemble_mean = target[name].mean("sample")
                gen_ensemble_mean = gen[name].mean("sample")
                # compute R2 values
                if ref is not None:
                    r2_target = get_r2(target_ensemble_mean, ref.mean)
                    r2_gen = get_r2(gen_ensemble_mean, ref.mean)
                    metrics[f"r2/{name}_target"] = r2_target
                    metrics[f"r2/{name}_gen"] = r2_gen
                    target_label = f"target R2: {r2_target:.2f}"
                    gen_label = f"gen R2: {r2_gen:.2f}"
                else:
                    target_label = "target"
                    gen_label = "gen"
                plot_mean_and_samples(
                    ax, target[name], target_label, color="orange", plot_samples=False
                )
                plot_mean_and_samples(ax, gen[name], gen_label)

            ax.set_title(f"{name}")
            ax.set_ylabel(f"{long_name} [{units}]")
            ax.legend()
            fig.tight_layout()
            plots[name] = fig
        if len(label) > 0:
            label = label + "/"
        logs = {}
        logs.update({f"{label}{name}": plots[name] for name in plots.keys()})
        logs.update({f"{label}{name}": metrics[name] for name in metrics.keys()})
        return logs

    def get_dataset(self) -> xr.Dataset:
        gathered = self._get_gathered_means()
        if gathered is None:
            return xr.Dataset()
        target, gen = gathered
        return xr.concat(
            [
                target.expand_dims({"source": ["target"]}),
                gen.expand_dims({"source": ["prediction"]}),
            ],
            dim="source",
        )


class GlobalMeanAnnualAggregator:
    def __init__(
        self,
        ops: GriddedOperations,
        timestep: datetime.timedelta,
        variable_metadata: Mapping[str, VariableMetadata] | None = None,
    ):
        self._area_weighted_mean_dict = ops.area_weighted_mean_dict
        self.timestep = timestep
        self.variable_metadata = variable_metadata or {}
        self._datasets: list[xr.Dataset] | None = None

    @torch.no_grad()
    def record_batch(self, time: xr.DataArray, data: TensorMapping):
        """Record a batch of data for computing time variability statistics."""
        data_area_mean = {
            name: tensor.cpu()
            for name, tensor in self._area_weighted_mean_dict(data).items()
        }
        ds = to_dataset(data_area_mean, time)

        # must keep a separate dataset for each sample to avoid averaging across
        # samples when we groupby year
        if self._datasets is None:
            self._datasets = []
            for i_sample in range(ds.sizes["sample"]):
                self._datasets.append(
                    ds.isel(sample=i_sample)
                    .groupby(ds["valid_time"].isel(sample=i_sample).dt.year)
                    .sum()
                )
        else:
            for i_sample in range(ds.sizes["sample"]):
                self._datasets[i_sample] = _add_dataarray(
                    self._datasets[i_sample],
                    ds.isel(sample=i_sample)
                    .groupby(ds["valid_time"].isel(sample=i_sample).dt.year)
                    .sum(),
                )

    def get_gathered_means(self) -> xr.Dataset | None:
        """
        Gather the mean data across all processes.

        Returns:
            The mean dataset, or None if this is not the root rank.
        """
        if self._datasets is None:
            raise ValueError("No data has been recorded yet.")
        dist = Distributed.get_instance()
        data = xr.concat(self._datasets, dim="sample")
        if dist.world_size > 1:
            data = _gather_sample_datasets(dist, data)
        if data is None:
            return None  # we are not root rank
        # filter out data with insufficient samples
        min_samples = _get_min_samples(self.timestep)
        data = data.where(data["counts"] > min_samples, drop=True)
        data = data / data["counts"]
        # ensure the 'year' coordinate has no jumps, filling in with NaNs as needed
        if data.sizes["year"] > 0:
            min_year = data["year"].min()
            max_year = data["year"].max()
            years = np.arange(min_year, max_year + 1, dtype=data.year.dtype)
            data = data.reindex(year=years)
        return data

    @torch.no_grad()
    def get_logs(self, label: str) -> dict[str, Any]:
        ds = self.get_gathered_means()
        if ds is None:  # not the root rank
            return {}
        plots = {}

        for name in ds.data_vars.keys():
            if name == "counts":
                continue

            if name in self.variable_metadata:
                long_name = self.variable_metadata[name].long_name
                units = self.variable_metadata[name].units
            else:
                long_name = name
                units = "unknown units"

            fig = Figure()  # create directly for cleanup when it leaves scope
            ax = fig.add_subplot(1, 1, 1)  # Add an axes to the figure
            if ds.sizes["year"] > 1:
                plot_mean_and_samples(ax, ds[name], "ensemble mean")
            ax.set_title(f"{name}")
            ax.set_ylabel(f"{long_name} [{units}]")
            ax.legend()
            fig.tight_layout()
            plots[name] = fig

        if len(label) > 0:
            label = label + "/"

        logs = {f"{label}{name}": plot for name, plot in plots.items()}
        return logs

    def get_dataset(self) -> xr.Dataset:
        gathered = self.get_gathered_means()
        if gathered is None:
            return xr.Dataset()
        return gathered


@dataclasses.dataclass
class VariableReferenceData:
    mean: xr.DataArray
    min: xr.DataArray
    max: xr.DataArray


def process_monthly_reference(
    monthly_reference_data: xr.Dataset,
    area_weighted_mean: Callable[[torch.Tensor], torch.Tensor],
    name: str,
) -> VariableReferenceData:
    ref_global_mean = xr.DataArray(
        area_weighted_mean(torch.as_tensor(monthly_reference_data[name].values)),
        dims=monthly_reference_data[name].dims[:-2],
        coords={"time": monthly_reference_data[name].coords["time"]},
    )
    valid_time_0 = monthly_reference_data.valid_time.isel(sample=0)
    for i in range(1, monthly_reference_data.sizes["sample"]):
        valid_time_i = monthly_reference_data.valid_time.isel(sample=i)
        if not valid_time_0.equals(valid_time_i):
            raise ValueError("All time axes must be the same")
    # need to select just one time axis so we don't lose sample dim
    ref_annual_coarsened = (ref_global_mean * monthly_reference_data["counts"]).groupby(
        valid_time_0.dt.year
    ).sum() / monthly_reference_data["counts"].groupby(valid_time_0.dt.year).sum()
    return VariableReferenceData(
        mean=ref_annual_coarsened.mean("sample"),
        min=ref_annual_coarsened.min("sample"),
        max=ref_annual_coarsened.max("sample"),
    )


def _add_dataarray(da1: xr.DataArray, da2: xr.DataArray):
    """
    Perform dataarray addition, assuming any missing year indices
    have zero values.
    """
    union_index = np.union1d(da1.year.values, da2.year.values)
    da1 = da1.reindex(year=union_index, fill_value=0)
    da2 = da2.reindex(year=union_index, fill_value=0)
    return da1 + da2


def get_r2(da: xr.DataArray, reference: xr.DataArray) -> float:
    """Compute the R2 value of the target compared to the reference."""
    ref_data = reference.sel(year=da.year)
    SS_ref = np.sum((ref_data.values - np.mean(ref_data.values)) ** 2)
    SS_pred = np.sum((da - ref_data).values ** 2)
    return float(1 - SS_pred / SS_ref)


def _gather_sample_datasets(
    dist: Distributed, dataset: xr.Dataset
) -> xr.Dataset | None:
    """
    Gather the dataset across all processes, concatenating on the sample dimension.

    Assumes all dataset variables have the same dimensions and shape, and that the
    first dimension is "sample".
    """
    # collect all data into one torch.Tensor for gathering, sort for determinism
    names = sorted(list(dataset.data_vars))
    tensor = torch.cat(
        [torch.asarray(np.expand_dims(dataset[name].values, axis=0)) for name in names],
        dim=0,
    ).to(get_device())
    years = torch.asarray(dataset.year.values).to(get_device())
    gathered_tensors = dist.gather_irregular(tensor)
    gathered_years = dist.gather_irregular(years)
    if gathered_tensors is None or gathered_years is None:
        return None
    datasets = []
    for tensor, years in zip(gathered_tensors, gathered_years):
        single_rank_dataset = xr.Dataset(
            {
                name: (["sample", "year"], tensor[i].cpu())
                for i, name in enumerate(names)
            },
            coords={"year": years.cpu()},
        )
        datasets.append(single_rank_dataset)
    # concat ranks along sample dim
    dataset_out = xr.concat(datasets, dim="sample")
    return dataset_out


@torch.no_grad()
def to_dataset(data: TensorMapping, time: xr.DataArray) -> xr.Dataset:
    """Convert a dictionary of data to an xarray dataset."""
    assert time.dims == ("sample", "time")  # must be consistent with this module
    data_vars = {}
    for name, tensor in data.items():
        data_vars[name] = (["sample", "time"], tensor)
    data_vars["counts"] = (
        ["sample", "time"],
        np.ones(shape=time.shape, dtype=np.float32),
    )
    return xr.Dataset(data_vars, coords={"valid_time": time})


def _get_min_samples(timestep: datetime.timedelta) -> int:
    steps_per_day = datetime.timedelta(days=1) // timestep
    return 362 * steps_per_day



"""This script downloads specified variables of the ERA5 dataset from
the NCAR Research Data Archive (RDA) and uploads them to Google Cloud Storage."""

import os
import shutil
import subprocess
import tempfile
import time
from dataclasses import dataclass
from datetime import datetime
from typing import List, Tuple

import click
import fsspec
import pandas as pd

MEAN_FLUX_CATEGORY_NAME = "e5.oper.fc.sfc.meanflux"
INVARIANT_CATEGORY_NAME = "e5.oper.invariant"
SURFACE_ANALYSIS_CATEGORY_NAME = "e5.oper.an.sfc"

NCAR_DATA = "https://data.rda.ucar.edu"
NCAR_STRATUS = "https://stratus.rda.ucar.edu"


@dataclass
class VariableFile:
    category: str
    variable_name: str
    start_time: datetime
    end_time: datetime

    @property
    def filename(self) -> str:
        start_str = self.start_time.strftime("%Y%m%d%H")
        end_str = self.end_time.strftime("%Y%m%d%H")
        return f"{self.category}.{self.variable_name}.ll025sc.{start_str}_{end_str}.nc"

    def get_rda_url(self, server=NCAR_DATA) -> str:
        year_month = self.start_time.strftime("%Y%m")
        prefix = f"{server}/ds633.0"
        return f"{prefix}/{self.category}/{year_month}/{self.filename}"


def assert_valid_start_time(category: str, start_time: datetime):
    """
    Assert that the category and times are valid.
    """
    if category == MEAN_FLUX_CATEGORY_NAME:
        assert start_time.hour == 6, "All files start at 06Z"
        assert start_time.day in (1, 16), "Files start on 1st or 16th of month"
    elif category == SURFACE_ANALYSIS_CATEGORY_NAME:
        assert start_time.hour == 0, "All files start at 00Z"
        assert start_time.day == 1, "Files start on 1st of month"
    elif category == INVARIANT_CATEGORY_NAME:
        assert start_time == datetime(1979, 1, 1, 0)


def get_date_list(
    category: str, start_time: datetime, n_files: int
) -> List[Tuple[datetime, datetime]]:
    """
    Get a list of dates for a given category and time range.
    """
    assert_valid_start_time(category, start_time)

    if category == INVARIANT_CATEGORY_NAME:
        date_list = [(start_time, start_time)]
    elif category == MEAN_FLUX_CATEGORY_NAME:
        times = pd.date_range(start_time, freq="SMS-16", periods=n_files + 1)
        date_list = list(zip(times[:-1], times[1:]))
    elif category == SURFACE_ANALYSIS_CATEGORY_NAME:
        start_times = pd.date_range(start_time, periods=n_files, freq="MS")
        end_times = pd.date_range(
            start_time.strftime("%Y-%m"), periods=n_files, freq="M"
        ) + pd.Timedelta("23h")
        date_list = list(zip(start_times, end_times))
    return date_list


def get_file_list(
    category: str, variable_name: str, start_time: datetime, n_files: int
) -> List[VariableFile]:
    """
    Get a list of files for a given category, variable, and time range.
    """
    date_list = get_date_list(category, start_time, n_files)
    file_list = []
    for start_date, end_date in date_list:
        file_list.append(VariableFile(category, variable_name, start_date, end_date))
    return file_list


def wget_file(url: str, local_path: str) -> None:
    subprocess.check_call(["wget", "-N", "-O", local_path, url])


def upload_files_to_gcs(local_dir: str, gcs_dir: str) -> None:
    subprocess.check_call(["gsutil", "-m", "cp", f"{local_dir}/*.nc", gcs_dir])


@click.command()
@click.argument("category")
@click.argument("variable_name")
@click.argument("start_time")
@click.argument("n_files", type=int)
@click.option(
    "--gcs-dir",
    default="gs://vcm-ml-scratch/oliwm/era5",
    help="Output path in GCS.",
    show_default=True,
)
@click.option(
    "--n-upload",
    default=8,
    help="Number of files over which to parallelize gsutil upload.",
    show_default=True,
)
def main(category, variable_name, start_time, n_files, gcs_dir, n_upload):
    """Ingest ERA5 data from NCAR RDA database to Google Cloud Storage.

    \b
    CATEGORY: one of 'e5.oper.fc.sfc.meanflux', 'e5.oper.invariant', 'e5.oper.an.sfc'
    NAME: variable name including code, e.g. '235_040_mtnlwrf'
    START_TIME: format YYYYMMDDHH
    N_FILES: the number of files to download starting at START_TIME
    """
    start_time = datetime.strptime(start_time, "%Y%m%d%H")
    file_list = get_file_list(category, variable_name, start_time, n_files)
    fs = fsspec.filesystem("gs")
    full_gcs_dir = f"{gcs_dir}/{category}/{variable_name}"

    with tempfile.TemporaryDirectory() as local_dir:
        scratch_dir = os.path.join(local_dir, "scratch")
        os.makedirs(scratch_dir)
        for i, file in enumerate(file_list):
            if fs.exists(f"{full_gcs_dir}/{file.filename}"):
                print(f"{file.filename} already exists on GCS, skipping download.")
            else:
                print(f"Downloading {file.filename} to local")
                local_file_path = os.path.join(scratch_dir, file.filename)
                # hacky retry logic to handle NCAR server issues
                try:
                    wget_file(file.get_rda_url(), local_file_path)
                except subprocess.CalledProcessError:
                    print("Trying again with different NCAR server")
                    time.sleep(15)
                    try:
                        wget_file(file.get_rda_url(NCAR_STRATUS), local_file_path)
                    except subprocess.CalledProcessError:
                        print("Trying original NCAR server again")
                        time.sleep(60)
                        wget_file(file.get_rda_url(), local_file_path)

            if i % n_upload == n_upload - 1 or i == len(file_list) - 1:
                # parallelize upload to GCS across n_upload files using gsutil -m
                if len(os.listdir(scratch_dir)) > 1:
                    print(f"Uploading the following files to GCS:")
                    print(os.listdir(scratch_dir))
                    upload_files_to_gcs(scratch_dir, full_gcs_dir)
                    shutil.rmtree(scratch_dir)
                    os.makedirs(scratch_dir)
                elif len(os.listdir(scratch_dir)) == 1:
                    # gsutil has different behavior for single file uploads
                    print(f"Uploading {os.listdir(scratch_dir)[0]} to GCS")
                    upload_path = f"{full_gcs_dir}/{os.listdir(scratch_dir)[0]}"
                    upload_files_to_gcs(scratch_dir, upload_path)
                    shutil.rmtree(scratch_dir)
                    os.makedirs(scratch_dir)
                else:
                    print("No files to upload to GCS")


def test_get_date_list():
    start_time = datetime(1979, 1, 1, 0)
    date_list = get_date_list(INVARIANT_CATEGORY_NAME, start_time, 1)
    assert len(date_list) == 1
    assert date_list[0] == (start_time, start_time)

    start_time = datetime(2020, 1, 1, 0)
    date_list = get_date_list(SURFACE_ANALYSIS_CATEGORY_NAME, start_time, 2)
    assert len(date_list) == 2
    assert date_list[0] == (start_time, datetime(2020, 1, 31, 23))
    assert date_list[1] == (datetime(2020, 2, 1, 0), datetime(2020, 2, 29, 23))

    start_time = datetime(2020, 1, 1, 6)
    date_list = get_date_list(MEAN_FLUX_CATEGORY_NAME, start_time, 2)
    assert len(date_list) == 2
    assert date_list[0] == (start_time, datetime(2020, 1, 16, 6))
    assert date_list[1] == (datetime(2020, 1, 16, 6), datetime(2020, 2, 1, 6))


if __name__ == "__main__":
    main()



import logging

import torch.utils.data
from torch.utils.data.distributed import DistributedSampler
from torch.utils.data.sampler import RandomSampler

from fme.ace.data_loading.batch_data import BatchData
from fme.ace.data_loading.dataloader import get_data_loader
from fme.ace.requirements import DataRequirements, PrognosticStateDataRequirements
from fme.core.dataset.config import (
    ConcatDatasetConfig,
    MergeDatasetConfig,
    MergeNoConcatDatasetConfig,
)
from fme.core.dataset.getters import (
    get_dataset,
    get_merged_datasets,
    get_xarray_dataset,
)
from fme.core.dataset.xarray import XarrayDataConfig, XarrayDataset
from fme.core.device import using_gpu
from fme.core.distributed import Distributed

from .batch_data import PrognosticState
from .config import DataLoaderConfig
from .gridded_data import GriddedData, InferenceGriddedData
from .inference import (
    ExplicitIndices,
    ForcingDataLoaderConfig,
    InferenceDataLoaderConfig,
    InferenceDataset,
)

logger = logging.getLogger(__name__)


class CollateFn:
    def __init__(self, horizontal_dims: list[str]):
        self.horizontal_dims = horizontal_dims

    def __call__(self, samples):
        return BatchData.from_sample_tuples(
            samples,
            horizontal_dims=self.horizontal_dims,
        )


def _get_sampler(
    dataset: torch.utils.data.Dataset,
    sample_with_replacement_dataset_size: int | None,
    train: bool,
) -> torch.utils.data.Sampler:
    dist = Distributed.get_instance()
    if sample_with_replacement_dataset_size is not None:
        local_sample_with_replacement_dataset_size = (
            sample_with_replacement_dataset_size // dist.world_size
        )
        sampler = torch.utils.data.RandomSampler(
            dataset,
            num_samples=local_sample_with_replacement_dataset_size,
            replacement=True,
        )
    elif dist.is_distributed():
        sampler = DistributedSampler(dataset, shuffle=train)
    else:
        sampler = RandomSampler(dataset) if train else None
    return sampler


def get_gridded_data(
    config: DataLoaderConfig,
    train: bool,
    requirements: DataRequirements,
) -> GriddedData:
    """
    Args:
        config: Parameters for the data loader.
        train: Whether loader is intended for training or validation data; if True,
            then data will be shuffled.
        requirements: Data requirements for the model.
    """
    n_timesteps_preloaded = config.time_buffer + requirements.n_timesteps
    # include requirements.n_timesteps - 1 steps of overlap so that no samples are
    # skipped at the boundaries of the preloaded timesteps
    start_every_n = config.time_buffer + 1
    dataset: torch.utils.data.Dataset
    if isinstance(config.dataset, XarrayDataConfig):
        dataset, properties = get_xarray_dataset(
            config.dataset,
            requirements.names,
            n_timesteps_preloaded,
        )
    elif isinstance(config.dataset, ConcatDatasetConfig):
        dataset, properties = get_dataset(
            config.dataset.concat,
            requirements.names,
            n_timesteps_preloaded,
            strict=config.dataset.strict,
        )
    elif isinstance(config.dataset, MergeDatasetConfig):
        dataset, properties = get_merged_datasets(
            config.dataset,
            requirements.names,
            n_timesteps_preloaded,
        )

    if config.time_buffer > 0:
        indices = range(len(dataset))[::start_every_n]
        dataset = torch.utils.data.Subset(dataset, indices)

    dist = Distributed.get_instance()

    sampler = _get_sampler(dataset, config.sample_with_replacement, train)

    if config.zarr_engine_used:
        # GCSFS and S3FS are not fork-safe, so we need to use forkserver
        # reading zarr with async from weka also requires forkserver
        mp_context = "forkserver"
        persistent_workers = True
    else:
        mp_context = None
        persistent_workers = False

    dist = Distributed.get_instance()
    batch_size = dist.local_batch_size(int(config.batch_size))

    dataloader = get_data_loader(
        dataset,
        batch_size=batch_size,
        n_window_timesteps=requirements.n_timesteps,
        time_buffer=config.time_buffer,
        num_workers=config.num_data_workers,
        sampler=sampler,
        shuffled=train,
        drop_last=True,
        pin_memory=using_gpu(),
        collate_fn=CollateFn(list(properties.horizontal_coordinates.dims)),
        multiprocessing_context=mp_context,
        persistent_workers=persistent_workers,
        prefetch_factor=config.prefetch_factor,
    )

    return GriddedData(
        loader=dataloader,
        properties=properties,
        sampler=sampler,
        modifier=config.augmentation.build_modifier(),
    )


def get_inference_data(
    config: InferenceDataLoaderConfig,
    total_forward_steps: int,
    window_requirements: DataRequirements,
    initial_condition: PrognosticState | PrognosticStateDataRequirements,
    surface_temperature_name: str | None = None,
    ocean_fraction_name: str | None = None,
) -> InferenceGriddedData:
    """
    Args:
        config: Parameters for the data loader.
        total_forward_steps: Total number of forward steps to take over the course of
            inference.
        window_requirements: Data requirements for the model.
        initial_condition: Initial condition for the inference, or a requirements object
            specifying how to extract the initial condition from the first batch of
            data
        surface_temperature_name: Name of the surface temperature variable. Can be
            set to None if no ocean temperature prescribing is being used.
        ocean_fraction_name: Name of the ocean fraction variable. Can be set to None
            if no ocean temperature prescribing is being used.

    Returns:
        A data loader for inference with coordinates and metadata.
    """
    dataset = InferenceDataset(
        config,
        total_forward_steps,
        window_requirements,
        surface_temperature_name,
        ocean_fraction_name,
    )
    properties = dataset.properties

    if config.zarr_engine_used:
        # GCSFS and S3FS are not fork-safe, so we need to use forkserver
        # persist workers since startup is slow
        mp_context = "forkserver"
        persistent_workers = True
    else:
        mp_context = None
        persistent_workers = False

    logging.info(f"Multiprocessing inference context: {mp_context or 'fork'}")

    # we roll our own batching in InferenceDataset, which is why batch_size=None below
    loader = torch.utils.data.DataLoader(
        dataset,
        batch_size=None,
        num_workers=config.num_data_workers,
        shuffle=False,
        pin_memory=using_gpu(),
        multiprocessing_context=mp_context,
        persistent_workers=persistent_workers,
    )
    gridded_data = InferenceGriddedData(
        loader=loader,
        initial_condition=initial_condition,
        properties=properties,
    )

    return gridded_data


def get_forcing_data(
    config: ForcingDataLoaderConfig,
    total_forward_steps: int,
    window_requirements: DataRequirements,
    initial_condition: PrognosticState,
    surface_temperature_name: str | None = None,
    ocean_fraction_name: str | None = None,
) -> InferenceGriddedData:
    """Return a GriddedData loader for forcing data based on the initial condition.
    This function determines the start indices for the forcing data based on the initial
    time in the provided initial condition.

    Args:
        config: Parameters for the forcing data loader.
        total_forward_steps: Total number of forward steps to take over the course of
            inference.
        window_requirements: Data requirements for the forcing data.
        initial_condition: Initial condition for the inference.
        surface_temperature_name: Name of the surface temperature variable. Can be
            set to None if no ocean temperature prescribing is being used.
        ocean_fraction_name: Name of the ocean fraction variable. Can be set to None
            if no ocean temperature prescribing is being used.

    Returns:
        A data loader for forcing data with coordinates and metadata.
    """
    initial_time = initial_condition.as_batch_data().time
    if initial_time.shape[1] != 1:
        raise NotImplementedError("code assumes initial time only has 1 timestep")
    if isinstance(config.dataset, XarrayDataConfig):
        available_times = XarrayDataset(
            config.dataset, window_requirements.names, window_requirements.n_timesteps
        ).all_times
    elif isinstance(config.dataset, MergeNoConcatDatasetConfig):
        # Some forcing variables may not be in the first dataset,
        # use an empty data requirements to get all times
        if isinstance(config.dataset.merge[0], XarrayDataConfig):
            available_times = XarrayDataset(
                config.dataset.merge[0],
                names=[],
                n_timesteps=window_requirements.n_timesteps,
            ).all_times
        else:
            raise ValueError("Forcing data cannot be concatenated.")
    start_time_indices = []
    for time in initial_time.values[:, 0]:
        start_time_indices.append(available_times.get_loc(time))
    inference_config = config.build_inference_config(
        start_indices=ExplicitIndices(start_time_indices)
    )
    return get_inference_data(
        config=inference_config,
        total_forward_steps=total_forward_steps,
        window_requirements=window_requirements,
        initial_condition=initial_condition,
        surface_temperature_name=surface_temperature_name,
        ocean_fraction_name=ocean_fraction_name,
    )



import os

import torch

from fme.core.device import get_device
from fme.core.testing import validate_tensor

from .sfnonet import SphericalFourierNeuralOperatorNet

DIR = os.path.abspath(os.path.dirname(__file__))


def test_sfnonet_output_is_unchanged():
    torch.manual_seed(0)
    input_channels = 2
    output_channels = 3
    img_shape = (9, 18)
    n_samples = 4
    device = get_device()
    model = SphericalFourierNeuralOperatorNet(
        params=None,
        embed_dim=16,
        num_layers=2,
        img_shape=img_shape,
        in_chans=input_channels,
        out_chans=output_channels,
    ).to(device)
    # must initialize on CPU to get the same results on GPU
    x = torch.randn(n_samples, input_channels, *img_shape).to(device)
    with torch.no_grad():
        output = model(x)
    assert output.shape == (n_samples, output_channels, *img_shape)
    validate_tensor(
        output,
        os.path.join(DIR, "testdata/test_sfnonet_output_is_unchanged.pt"),
    )



import time

import cftime
import click
import numpy as np
import xarray as xr
from compute_ocean_dataset import OceanDatasetComputationConfig, OceanDatasetConfig
from dask.diagnostics import ProgressBar
from ocean_emulators.preprocessing import spatially_filter
from xgcm import Grid

LATENT_HEAT_OF_VAPORIZATION = 2.5e6  # J/kg
_LATENT_HEAT_VAPORIZATION_0_C = 2.5e6  # J/kg
_SPECIFIC_ENTHALPY_LIQUID = 4185.5
_SPECIFIC_ENTHALPY_VAP0R = 1846
_FREEZING_TEMPERATURE = 273.15


def latent_heat_vaporization(T):
    return _LATENT_HEAT_VAPORIZATION_0_C + (
        _SPECIFIC_ENTHALPY_LIQUID - _SPECIFIC_ENTHALPY_VAP0R
    ) * (T - _FREEZING_TEMPERATURE)


def open_ocean_dataset(config):
    ds = xr.open_mfdataset(config.ocean_dataset_nc_files)
    ds_monthly_thickness = xr.open_mfdataset(
        config.ocean_dataset_monthly_layer_thickness_files
    )
    upsampled = ds_monthly_thickness.interp(
        Time=ds.Time.values, method="nearest", kwargs={"fill_value": "extrapolate"}
    )
    ds["layer_thickness"] = upsampled[config.standard_names.ocean_layer_thickness]
    mesh = xr.open_dataset(config.nc_grid_path)
    ds["refBottomDepth"] = mesh.refBottomDepth
    wetmask = np.isnan(mesh.temperature.isel(Time=0))
    ds["wetmask"] = wetmask
    ds.wetmask.attrs["long_name"] = "wet mask"
    ds.wetmask.attrs["units"] = ""
    return ds


def open_ice_dataset(config):
    ds = xr.open_mfdataset(
        config.ice_dataset_nc_files,
        concat_dim=config.standard_names.time_dim,
        combine="nested",
    )
    return ds


def open_dataset(config):
    ocean = open_ocean_dataset(config)
    ice = open_ice_dataset(config)
    ds = xr.merge([ocean, ice])
    return ds


def get_no_spatially_filtered_variables(ds):
    return ds[["refBottomDepth", "wetmask", "layer_thickness"]]


def compute_surface_downward_heat_flux(ds, ocean_stream_prefix="timeCustom_avg_"):
    """In CM4, hfds is defined as
    Surface ocean heat flux from
    SW+LW+latent+sensible+masstransfer+frazil+seaice_melt_heat.
    In E3SM, we compute this as the sum of the following variables:
    shortWaveHeatFlux, longWaveHeatFluxDown
    evaporationFlux, sensibleHeatFlux,
    """
    ocean_downward_heat_flux = (
        ds[ocean_stream_prefix + "shortWaveHeatFlux"]
        + ds[ocean_stream_prefix + "longWaveHeatFluxDown"]
        + ds[ocean_stream_prefix + "evaporationFlux"] * LATENT_HEAT_OF_VAPORIZATION
        + ds[ocean_stream_prefix + "sensibleHeatFlux"]
    )
    return ocean_downward_heat_flux


def compute_vertical_coarsening(ds: xr.Dataset, config, variable_names) -> xr.Dataset:
    standard_names = config.standard_names
    target_vertical_levels = np.array(config.ocean_vertical_target_interface_levels)
    var2D = []
    var3D = []
    for var in variable_names:
        if ds[var].dims == (
            standard_names.time_dim,
            standard_names.latitude_dim,
            standard_names.longitude_dim,
        ):
            var2D.append(var)
        elif ds[var].dims == (
            standard_names.time_dim,
            standard_names.vertical_dim,
            standard_names.latitude_dim,
            standard_names.longitude_dim,
        ):
            var3D.append(var)
        else:
            print(f"Skipping {var} since it is a coordinate variable.")
    ds_2d = ds[var2D].copy()
    lev = ds.refBottomDepth
    lev_outer = np.zeros(len(lev))
    lev_outer[0] = lev.values[0] * 0.5
    lev_outer[1:] = 0.5 * (lev.values[0:-1] + lev.values[1:])
    lev_outer = np.append(lev_outer, lev.values[-1] * 2.0 - lev_outer[-1])
    ds = ds.rename_dims({config.standard_names.vertical_dim: "lev"})
    ds = ds.assign_coords({"lev": lev, "lev_outer": lev_outer})
    grid = Grid(
        ds,
        coords={"Z": {"center": "lev", "outer": "lev_outer"}},
        boundary="fill",
        periodic=False,
        autoparse_metadata=False,
    )
    dz = ds["layer_thickness"]
    ds_extensive = ds[var3D] * dz
    ds_extensive = ds_extensive.assign_coords({"lev_outer": lev_outer})
    ds_extensive_regridded = xr.Dataset()
    for var in var3D:
        ds_extensive_regridded[var] = grid.transform(
            ds_extensive[var],
            "Z",
            target_vertical_levels,
            target_data=ds.lev_outer,
            method="conservative",
        )
    ds_extensive_regridded = ds_extensive_regridded.rename({"lev_outer": "lev"})
    dz_regridded = xr.DataArray(
        np.diff(target_vertical_levels),
        dims=["lev"],
        coords={"lev": ds_extensive_regridded.lev},
    )
    ds_regridded = ds_extensive_regridded / dz_regridded
    ds_regridded = ds_regridded.assign_coords(dz=dz_regridded)
    for co_name, co in ds.coords.items():
        if "lev" not in co.dims:
            ds_regridded = ds_regridded.assign_coords({co_name: co})
    ds_regridded = ds_regridded.drop_vars("lev_outer")
    ds_regridded = ds_regridded.drop_vars("lev").rename_dims(
        {"lev": config.standard_names.vertical_dim}
    )
    ds_regridded = ds_regridded.drop_vars("dz")
    ds_regridded = xr.merge([ds_regridded, ds_2d])
    return ds_regridded


def rename_and_split_dataset_to_individual_layer(ds, config):
    rename = config.renaming
    ds = ds.rename(rename)
    for var in ds.variables:
        if config.standard_names.vertical_dim in ds[var].dims:
            for i in range(len(ds[config.standard_names.vertical_dim])):
                ds[f"{var}_{i}"] = ds[var].isel({config.standard_names.vertical_dim: i})
            ds = ds.drop_vars(var)
    return ds


def get_vertical_coarsened_wetmask(ds, config):
    standard_names = config.standard_names
    for var in ds.variables:
        if standard_names.sea_water_potential_temperature in var:
            wetmask = np.isnan(ds[var].isel(time=0))
            current_level = var.split("_")[-1]
            ds["wetmask_" + current_level] = wetmask

    return ds


def reindex_time_with_xtime(ds, config):
    timestamps = []
    for i in range(len(ds.xtime)):
        date_string = ds.xtime.values[i].decode("utf-8")
        date_part, time_part = date_string.split("_")
        year, month, day = date_part.split("-")
        hour, minute, _ = time_part.split(":")
        timestamps.append(
            cftime.DatetimeNoLeap(
                int(year), int(month), int(day), int(hour), int(minute)
            )
        )
        ds = ds.assign_coords({config.standard_names.time_dim: timestamps})
        return ds


def construct_lazy_dataset(
    config: OceanDatasetComputationConfig,
) -> xr.Dataset:
    start = time.time()
    standard_names = config.standard_names
    print(f"Opening dataset...")
    ds = open_dataset(config)
    print(f"Dataset opened in {time.time() - start:.2f} s total.")
    print(f"Input dataset size is {ds.nbytes / 1e9} GB")
    ds = reindex_time_with_xtime(ds, config)
    spatially_filtered_variables = list(config.renaming.keys())
    if config.compute_e3sm_surface_downward_heat_flux:
        ds["sfc_heat_flux"] = compute_surface_downward_heat_flux(ds)
        spatially_filtered_variables.append("sfc_heat_flux")
    ds_unfiltered = get_no_spatially_filtered_variables(ds)
    ds_filtered = spatially_filter(
        ds[spatially_filtered_variables],
        ds["wetmask"],
        depth_dim=standard_names.vertical_dim,
        y_dim=standard_names.latitude_dim,
        x_dim=standard_names.longitude_dim,
    )
    ds = xr.merge([ds_unfiltered, ds_filtered])
    ds = compute_vertical_coarsening(ds, config, spatially_filtered_variables)
    ds = rename_and_split_dataset_to_individual_layer(ds, config)
    ds = get_vertical_coarsened_wetmask(ds, config)
    renamed_standard_names = config.standard_names
    renamed_standard_names.time_dim = "time"
    chunks = config.chunking.get_chunks(renamed_standard_names)
    ds = ds.chunk(chunks).astype(np.float32)
    ds.attrs["history"] = (
        "Dataset computed by full-model/scripts"
        "/compute_ocean_dataset_e3smv2.py"
        f" script, using inputs from the following: {config.ocean_dataset_nc_files}"
        f" and {config.ice_dataset_nc_files}."
    )
    return ds


@click.command()
@click.option("--config", help="Path to dataset configuration YAML file.")
@click.option("-o", "--output", help="URL to write output to.")
@click.option("--debug", is_flag=True, help="Print metadata instead of writing output.")
@click.option("--subsample", is_flag=True, help="Subsample the data before writing.")
@click.option("--n-workers", default=4, help="Number of Dask workers.")
@click.option("--ranks", default=16, help="total number of available ranks")
@click.option("--rank", default=0, help="rank of job")
def main(
    config,
    output,
    debug,
    subsample,
    n_workers,
    ranks,
    rank,
):
    xr.set_options(keep_attrs=True)
    # _ = Client(n_workers=n_workers, threads_per_worker=2)
    config = OceanDatasetConfig.from_file(config).dataset_computation
    ds = construct_lazy_dataset(config)
    if subsample:
        ds = ds.isel(time=slice(10, 13))
    print(f"Output dataset size is {ds.nbytes / 1e9} GB")
    if debug:
        with xr.set_options(display_max_rows=500):
            print(ds)
    else:
        ds.partition.initialize_store(output)
        for i in range(config.n_split):
            print(f"Writing segment {i + 1} / {config.n_split}")
            with ProgressBar():
                ds.partition.write(
                    output, config.n_split, ["time"], i, collect_variable_writes=True
                )


if __name__ == "__main__":
    main()



import numpy as np
import pytest
import torch
import xarray as xr

from fme.ace.data_loading.batch_data import BatchData, PairedData
from fme.core.device import get_device


def assert_metadata_equal(a: BatchData, b: BatchData):
    assert a.horizontal_dims == b.horizontal_dims


def get_batch_data(
    names: list[str],
    n_samples: int,
    n_times: int,
    horizontal_dims: list[str],
    n_lat: int = 8,
    n_lon: int = 16,
):
    device = get_device()
    return BatchData(
        data={
            name: torch.randn(n_samples, n_times, n_lat, n_lon, device=device)
            for name in names
        },
        time=xr.DataArray(np.random.rand(n_samples, n_times), dims=["sample", "time"]),
        horizontal_dims=horizontal_dims,
    )


@pytest.mark.parametrize(
    "names, prognostic_names",
    [
        pytest.param(["prog"], ["prog"], id="all prognostic"),
        pytest.param(["prog", "forcing"], ["prog"], id="some prognostic"),
        pytest.param(["forcing1", "forcing2"], [], id="no prognostic"),
    ],
)
@pytest.mark.parametrize("n_ic_timesteps", [1, 2])
def test_get_start(names: list[str], prognostic_names: list[str], n_ic_timesteps: int):
    n_samples = 2
    n_times = 5
    n_lat = 8
    n_lon = 16
    horizontal_dims = ["lat", "lon"]
    batch_data = get_batch_data(
        names=names,
        n_samples=n_samples,
        n_times=n_times,
        horizontal_dims=horizontal_dims,
        n_lat=n_lat,
        n_lon=n_lon,
    )
    start = batch_data.get_start(prognostic_names, n_ic_timesteps).as_batch_data()
    assert_metadata_equal(start, batch_data)
    assert start.time.equals(batch_data.time.isel(time=slice(0, n_ic_timesteps)))
    assert set(start.data.keys()) == set(prognostic_names)
    for name in prognostic_names:
        assert start.data[name].shape == (n_samples, n_ic_timesteps, n_lat, n_lon)
        np.testing.assert_allclose(
            start.data[name].cpu().numpy(),
            batch_data.data[name][:, :n_ic_timesteps, ...].cpu().numpy(),
        )


@pytest.mark.parametrize(
    "names, prognostic_names",
    [
        pytest.param(["prog"], ["prog"], id="all prognostic"),
        pytest.param(["prog", "forcing"], ["prog"], id="some prognostic"),
        pytest.param(["forcing1", "forcing2"], [], id="no prognostic"),
    ],
)
@pytest.mark.parametrize("n_ic_timesteps", [1, 2])
def test_get_end(names: list[str], prognostic_names: list[str], n_ic_timesteps: int):
    n_samples = 2
    n_times = 5
    n_lat = 8
    n_lon = 16
    horizontal_dims = ["lat", "lon"]
    batch_data = get_batch_data(
        names=names,
        n_samples=n_samples,
        n_times=n_times,
        horizontal_dims=horizontal_dims,
        n_lat=n_lat,
        n_lon=n_lon,
    )
    end = batch_data.get_end(prognostic_names, n_ic_timesteps).as_batch_data()
    assert_metadata_equal(end, batch_data)
    assert end.time.equals(batch_data.time.isel(time=slice(-n_ic_timesteps, None)))
    assert set(end.data.keys()) == set(prognostic_names)
    for name in prognostic_names:
        assert end.data[name].shape == (n_samples, n_ic_timesteps, n_lat, n_lon)
        np.testing.assert_allclose(
            end.data[name].cpu().numpy(),
            batch_data.data[name][:, -n_ic_timesteps:, ...].cpu().numpy(),
        )


@pytest.mark.parametrize(
    "names, prepend_names",
    [
        pytest.param(["prog"], ["prog"], id="all prepended"),
        pytest.param(["prog", "forcing"], ["prog"], id="some prepended"),
    ],
)
@pytest.mark.parametrize("n_ic_timesteps", [1, 2])
def test_prepend(names: list[str], prepend_names: list[str], n_ic_timesteps: int):
    n_samples = 2
    n_times = 5
    n_lat = 8
    n_lon = 16
    horizontal_dims = ["lat", "lon"]
    batch_data = get_batch_data(
        names=names,
        n_samples=n_samples,
        n_times=n_times,
        horizontal_dims=horizontal_dims,
        n_lat=n_lat,
        n_lon=n_lon,
    )
    start_data = batch_data.get_start(prepend_names, n_ic_timesteps)
    prepended = batch_data.prepend(start_data)
    start_batch_data = start_data.as_batch_data()
    assert_metadata_equal(prepended, batch_data)
    assert prepended.time.isel(time=slice(n_ic_timesteps, None)).equals(batch_data.time)
    assert set(prepended.data.keys()) == set(names)
    for name in names:
        np.testing.assert_allclose(
            prepended.data[name][:, n_ic_timesteps:, ...].cpu().numpy(),
            batch_data.data[name].cpu().numpy(),
        )
    for name in prepend_names:
        np.testing.assert_allclose(
            prepended.data[name][:, :n_ic_timesteps, ...].cpu().numpy(),
            start_batch_data.data[name].cpu().numpy(),
        )
    assert prepended.time.shape == (n_samples, n_times + n_ic_timesteps)
    for name in set(names) - set(prepend_names):
        assert np.all(
            np.isnan(prepended.data[name][:, :n_ic_timesteps, ...].cpu().numpy())
        )


def test_paired_data_forcing_target_data():
    n_samples = 2
    n_times = 5
    n_lat = 8
    n_lon = 16
    horizontal_dims = ["lat", "lon"]
    target_data = get_batch_data(
        names=["foo", "bar"],
        n_samples=n_samples,
        n_times=n_times,
        horizontal_dims=horizontal_dims,
        n_lat=n_lat,
        n_lon=n_lon,
    )
    gen_data = get_batch_data(
        names=["bar"],
        n_samples=n_samples,
        n_times=n_times,
        horizontal_dims=horizontal_dims,
        n_lat=n_lat,
        n_lon=n_lon,
    )
    gen_data.time = target_data.time
    paired_data = PairedData.from_batch_data(prediction=gen_data, reference=target_data)
    assert paired_data.forcing == {"foo": target_data.data["foo"]}
    assert paired_data.target == {"bar": target_data.data["bar"]}



from fme.core.cli import get_parser

from .inference import main

parser = get_parser()
parser.add_argument(
    "--segments",
    type=int,
    default=None,
    help="If provided, number of times to repeat the inference in time, saving each "
    "segment in a separate folder labeled as 'segment_0000', 'segment_0001' etc. "
    "WARNING: this feature is experimental and its API is subject to change.",
)
args = parser.parse_args()
main(args.yaml_config, segments=args.segments, override_dotlist=args.override)



import pytest
import torch

from fme.core.prescriber import Prescriber, PrescriberConfig


def test_prescriber_config_build_raises_value_error():
    """Test that error is raised if name not in in_names and out_names."""
    config = PrescriberConfig(prescribed_name="a", mask_name="b", mask_value=1)
    with pytest.raises(ValueError):
        config.build(in_names=["a"], out_names=["c"])


def test_prescriber():
    """Test that the prescriber overwrites the generated data in the masked region."""
    prescriber = Prescriber(prescribed_name="a", mask_name="mask", mask_value=0)
    data = {
        "a": torch.rand(2, 4, 4),
        "b": torch.rand(2, 4, 4),
        "mask": torch.ones(2, 4, 4),
    }
    target = {
        "a": torch.rand(2, 4, 4),
        "b": torch.rand(2, 4, 4),
    }
    data["mask"][:, :, 2:] = 0
    gen = {k: torch.rand_like(v) for k, v in target.items()}
    expected_gen = {k: v.clone() for k, v in gen.items()}
    expected_gen["a"][:, :, 2:] = target["a"][:, :, 2:]
    assert not torch.allclose(gen["a"], expected_gen["a"])
    prescribed_gen = prescriber(data, gen, target)
    for name in gen:
        torch.testing.assert_close(prescribed_gen[name], expected_gen[name])
    # non-integer valued mask
    prescriber = Prescriber(prescribed_name="a", mask_name="mask", mask_value=1)
    data["mask"] = torch.zeros(2, 4, 4, dtype=torch.float32) + 0.1
    data["mask"][:, :, 2:] = 0.7
    prescribed_gen = prescriber(data, gen, target)
    for name in gen:
        torch.testing.assert_close(prescribed_gen[name], expected_gen[name])


def test_prescriber_interpolate():
    prescriber = Prescriber(
        prescribed_name="a", mask_name="mask", mask_value=1, interpolate=True
    )
    data = {
        "a": torch.zeros(2, 4, 4),
        "b": torch.ones(2, 4, 4) * 4.0,
        "mask": torch.ones(2, 4, 4) * 0.25,
    }
    target = {
        "a": torch.ones(2, 4, 4) * 4.0,
        "b": torch.zeros(2, 4, 4),
    }
    prescribed_gen = prescriber(data, data, target)
    torch.testing.assert_close(prescribed_gen["a"], torch.ones(2, 4, 4))
    # check that the other variable is not changed
    torch.testing.assert_close(prescribed_gen["b"], torch.ones(2, 4, 4) * 4.0)



import collections
import contextlib
import random
import string
from collections.abc import Mapping
from typing import Any, Literal

from fme.core import wandb
from fme.core.distributed import Distributed


class MockWandB:
    def __init__(self):
        self._enabled = False
        self._configured = False
        self._logs: dict[int, dict[str, Any]] = collections.defaultdict(dict)
        self._last_step = 0
        self._id: str | None = None

    def configure(self, log_to_wandb: bool):
        dist = Distributed.get_instance()
        self._enabled = log_to_wandb and dist.is_root()
        self._configured = True

    def init(
        self,
        resumable: bool = False,
        experiment_dir: str | None = None,
        **kwargs,
    ):
        if not self._configured:
            raise RuntimeError(
                "must call WandB.configure before WandB init can be called"
            )
        if self._enabled:
            if resumable:
                if experiment_dir is None:
                    raise ValueError(
                        "must provide `experiment_dir` when `resumable` is True"
                    )
                else:
                    wandb.init_wandb_with_resumption(
                        experiment_dir,
                        direct_access=False,
                        wandb_init=self._wandb_init,
                        wandb_id=self.get_id,
                        **kwargs,
                    )
            else:
                self._wandb_init(resume="never", **kwargs)

    def _wandb_init(
        self, resume: Literal["must", "never"], id: str | None = None, **kwargs
    ):
        """
        Mocks the `wandb.init` behavior, specifically around initializing
        a run with `resume` and `id`.
        See https://docs.wandb.ai/guides/runs/resuming/.
        """
        if resume == "must":
            if id is None:
                raise ValueError("resume='must' and id is None")
            else:
                if id != self._id:
                    raise ValueError("resume='must' and id does not match previous id")
        else:
            if id is not None:
                raise ValueError("resume='never' and id is not None")
            else:
                if self._id is not None:
                    raise ValueError(
                        "resume='never' and id is None but previous id exists"
                    )
            self._id = _mock_wandb_id()

    def get_id(self) -> str:
        if self._id is None:
            raise ValueError("mock wandb id is None")
        return self._id

    def set_id(self, id: str):
        self._id = id

    def watch(self, modules):
        if self._enabled:
            # wandb.watch(modules)
            pass

    def log(self, data: Mapping[str, Any], step: int, sleep=None):
        if step < self._last_step:
            raise ValueError(
                f"step {step} is less than last step {self._last_step}, "
                "steps must be logged in order"
            )
        self._last_step = step
        # sleep arg is ignored since we don't want to sleep in tests
        if self._enabled:
            self._logs[step].update(data)

    def get_logs(self) -> list[dict[str, Any]]:
        if len(self._logs) == 0:
            return []
        n_logs = max(self._logs.keys())
        return_value: list[dict[str, Any]] = [dict() for _ in range(n_logs + 1)]
        for step, log in self._logs.items():
            return_value[step] = log
        return return_value

    def Image(self, *args, **kwargs) -> wandb.Image:
        return wandb.Image(*args, direct_access=False, **kwargs)

    def Video(self, *args, **kwargs) -> wandb.Video:
        return wandb.Video(*args, direct_access=False, **kwargs)

    def Table(self, *args, **kwargs) -> wandb.Table:
        return wandb.Table(*args, direct_access=False, **kwargs)

    def Histogram(self, *args, **kwargs) -> wandb.Histogram:
        return wandb.Histogram(*args, direct_access=False, **kwargs)

    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def configured(self) -> bool:
        return self._configured


@contextlib.contextmanager
def mock_wandb():
    """
    Mock the distributed singleton to return a MockDistributed object.

    This is useful for testing that metrics are reduced across processes.

    It will make it so that when any tensor is reduced, it is filled with
    the given fill_value, which can be checked for in tests.
    """
    original = wandb.singleton
    wandb.singleton = MockWandB()  # type: ignore
    try:
        yield wandb.singleton
    finally:
        wandb.singleton = original


def _mock_wandb_id(n_chars: int = 8) -> str:
    return "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(n_chars)
    )



import pytest
import torch

from fme.core import metrics
from fme.core.device import get_device
from fme.core.gridded_ops import GriddedOperations, LatLonOperations
from fme.core.loss import (
    AreaWeightedMSELoss,
    CRPSLoss,
    EnergyScoreLoss,
    GlobalMeanLoss,
    LossConfig,
    VariableWeightingLoss,
    WeightedMappingLoss,
    WeightedMappingLossConfig,
    _construct_weight_tensor,
)
from fme.core.normalizer import StandardNormalizer
from fme.core.packer import Packer


@pytest.mark.parametrize("global_mean_type", [None, "LpLoss"])
def test_loss_builds_and_runs(global_mean_type):
    config = LossConfig(global_mean_type=global_mean_type)
    area = torch.randn(10, 1, device=get_device()).broadcast_to(size=(10, 10))
    loss = config.build(
        reduction="mean",
        gridded_operations=LatLonOperations(area),
    )
    x = torch.randn(10, 10, 10, 10, 10, device=get_device())
    y = torch.randn(10, 10, 10, 10, 10, device=get_device())
    result = loss(x, y)
    assert result.shape == ()
    assert isinstance(result, torch.Tensor)


def test_spectral_energy_score():
    torch.manual_seed(0)
    DEVICE = get_device()
    n_lat, n_lon = 16, 32
    pred = torch.rand(10000, 2, n_lat, n_lon, device=DEVICE)
    target = torch.rand(10000, 2, n_lat, n_lon, device=DEVICE)
    sht = LatLonOperations(torch.ones((n_lat, n_lon), device=DEVICE)).get_real_sht(
        grid="legendre-gauss"
    )
    spectral_energy_score_loss = EnergyScoreLoss(sht=sht)
    crps_loss = CRPSLoss(alpha=0.95)
    score = spectral_energy_score_loss(pred, target)
    crps = crps_loss(pred, target)

    n_lat2, n_lon2 = 32, 64
    pred = torch.rand(10000, 2, n_lat2, n_lon2, device=DEVICE)
    target = torch.rand(10000, 2, n_lat2, n_lon2, device=DEVICE)
    sht = LatLonOperations(torch.ones((n_lat2, n_lon2), device=DEVICE)).get_real_sht(
        grid="legendre-gauss"
    )
    spectral_energy_score_loss = EnergyScoreLoss(sht=sht)
    larger_domain_score = spectral_energy_score_loss(pred, target)
    torch.testing.assert_close(larger_domain_score, score, rtol=0.05, atol=0.0)
    torch.testing.assert_close(score, crps, rtol=0.5, atol=0.0)


def test_loss_of_zeros_is_variance():
    torch.manual_seed(0)
    config = LossConfig(global_mean_type=None)
    loss = config.build(
        reduction="mean",
        gridded_operations=LatLonOperations(torch.ones(10, 10)),
    )
    x = torch.zeros(10, 10, 10, 10, 10, device=get_device())
    y = torch.randn(10, 10, 10, 10, 10, device=get_device())
    result = loss(x, y)
    assert result.shape == ()
    assert isinstance(result, torch.Tensor)
    if str(get_device()).startswith("cuda"):
        tol = {"rtol": 1e-4, "atol": 1e-4}
    elif str(get_device()).startswith("mps"):
        tol = {"rtol": 1e-3, "atol": 1e-3}
    else:
        tol = {}
    torch.testing.assert_close(result, y.var(), **tol)


@pytest.mark.parametrize("global_mean_weight", [0.0, 1.0, 5.0])
def test_loss_of_zeros_is_one_plus_global_mean_weight(global_mean_weight: float):
    torch.manual_seed(0)
    config = LossConfig(
        global_mean_type="LpLoss", global_mean_weight=global_mean_weight
    )
    area = torch.randn(10, 1, device=get_device()).broadcast_to(size=(10, 10))
    loss = config.build(
        reduction="mean",
        gridded_operations=LatLonOperations(area),
    )
    x = torch.zeros(10, 10, 10, 10, 10, device=get_device())
    y = torch.randn(10, 10, 10, 10, 10, device=get_device())
    result = loss(x, y)
    assert result.shape == ()
    assert isinstance(result, torch.Tensor)
    expected = torch.tensor(1.0 + global_mean_weight)
    tol = (
        {"atol": 0.015, "rtol": 0.01}
        if str(get_device()).startswith("cuda")
        else {"atol": 0.01, "rtol": 0.0}
    )
    torch.testing.assert_close(result.cpu(), expected, **tol)


@pytest.mark.parametrize(
    "config",
    [
        LossConfig(type="AreaWeightedMSE"),
        LossConfig(global_mean_type="LpLoss"),
        LossConfig(
            type="EnsembleLoss", kwargs={"energy_score_weight": 1.0, "crps_weight": 0.0}
        ),
    ],
)
def test_loss_fails_when_gridded_operations_not_provided(
    config: LossConfig,
):
    with pytest.raises(ValueError):
        config.build(reduction="mean", gridded_operations=None)


def test_global_mean_loss():
    torch.manual_seed(0)
    area = torch.randn(10, 1, device=get_device()).broadcast_to(size=(10, 10))
    loss = GlobalMeanLoss(
        LatLonOperations(area).area_weighted_mean, loss=torch.nn.MSELoss()
    )
    x = torch.zeros(10, 10, 10, 10, 10, device=get_device())
    y = torch.randn(10, 10, 10, 10, 10, device=get_device())
    result = loss(x, y)
    assert result.shape == ()
    assert isinstance(result, torch.Tensor)

    def global_weighted_mean(tensor, area):
        return (tensor * area[None, None, None, :, :]).sum(dim=(-1, -2)) / area.sum()

    mse = torch.nn.MSELoss()
    expected = mse(global_weighted_mean(x, area), global_weighted_mean(y, area))
    torch.testing.assert_close(result, expected)


def test_area_weighted_mse():
    torch.manual_seed(0)
    x = torch.rand(10, 10).to(get_device())
    target = torch.rand(10, 10).to(get_device())
    area = torch.rand(10, 1, device=get_device()).broadcast_to(size=(10, 10))
    area_weighted_mse = AreaWeightedMSELoss(LatLonOperations(area).area_weighted_mean)
    result = area_weighted_mse(x, target)
    expected = metrics.weighted_mean(
        torch.nn.MSELoss(reduction="none")(x, target), weights=area, dim=(-2, -1)
    ).mean()
    torch.testing.assert_close(result, expected)


def test__construct_weight_tensor():
    out_names = ["a", "b", "c"]
    weights = {"a": 0.5, "c": 3.0}
    n_samples = 10
    nlat, nlon = 8, 4
    gen_data = torch.rand(n_samples, len(out_names), nlat, nlon, device=get_device())
    weight_tensor = _construct_weight_tensor(
        weights, out_names, n_dim=4, channel_dim=-3
    )
    weighted_gen_data = gen_data * weight_tensor
    assert weight_tensor.shape == (1, len(out_names), 1, 1)
    assert weighted_gen_data.shape == gen_data.shape
    for i, name in enumerate(out_names):
        if name in weights:
            assert torch.allclose(
                weighted_gen_data[:, i], weights[name] * gen_data[:, i]
            )
        else:
            assert torch.allclose(weighted_gen_data[:, i], gen_data[:, i])


@pytest.mark.parametrize("mean", [0.0, 1.0])
@pytest.mark.parametrize("scale", [1.0, 2.0])
def test_WeightedMappingLoss(mean, scale):
    loss = torch.nn.MSELoss()
    n_channels = 5
    packer = Packer([f"var_{i}" for i in range(n_channels)])
    out_names = [f"var_{i}" for i in range(n_channels)]
    normalizer = StandardNormalizer(
        means={name: torch.as_tensor(mean) for name in out_names},
        stds={name: torch.as_tensor(scale) for name in out_names},
    )
    mapping_loss = WeightedMappingLoss(
        loss,
        weights={},
        out_names=out_names,
        normalizer=normalizer,
    )
    x = torch.randn(
        15,
        n_channels,
        10,
        10,
    ).to(get_device(), dtype=torch.float)
    y = torch.randn(
        15,
        n_channels,
        10,
        10,
    ).to(get_device(), dtype=torch.float)
    x_mapping = {name: x[:, i, :, :] for i, name in enumerate(packer.names)}
    y_mapping = {name: y[:, i, :, :] for i, name in enumerate(packer.names)}
    assert torch.allclose(mapping_loss(x_mapping, y_mapping), loss(x, y) / scale**2)


def test_VariableWeightingLoss():
    weights = torch.tensor(
        [
            4.0,
            1.0,
        ]
    ).to(get_device())
    mse_loss = torch.nn.MSELoss()
    weighted_loss = VariableWeightingLoss(weights=weights, loss=mse_loss)

    x = torch.tensor([[1.0, 2.0]]).to(get_device())
    y = torch.tensor([[2, 2.5]]).to(get_device())

    weighted_result = weighted_loss(x, y)
    assert weighted_result == ((16 + 0.25) / 2.0)


def test_WeightedMappingLossConfig_no_weights():
    loss_config = LossConfig()
    n_channels = 5
    out_names = [f"var_{i}" for i in range(n_channels)]
    channel_dim = -3
    area = torch.tensor([])  # area not used by this config
    gridded_operations: GriddedOperations = LatLonOperations(area)
    mapping_loss_config = WeightedMappingLossConfig()
    loss = loss_config.build(reduction="mean", gridded_operations=gridded_operations)
    normalizer = StandardNormalizer(
        means={name: torch.as_tensor(0.0) for name in out_names},
        stds={name: torch.as_tensor(1.0) for name in out_names},
    )
    mapping_loss = mapping_loss_config.build(
        gridded_operations,
        out_names=out_names,
        channel_dim=channel_dim,
        normalizer=normalizer,
    )
    packer = Packer(out_names)

    x_mapping = {name: torch.randn(4, 5, 5).to(get_device()) for name in out_names}
    y_mapping = {name: torch.randn(4, 5, 5).to(get_device()) for name in out_names}
    x = packer.pack(x_mapping, axis=channel_dim)
    y = packer.pack(y_mapping, axis=channel_dim)

    assert loss(x, y) == mapping_loss(x_mapping, y_mapping)


def test_WeightedMappingLossConfig_weights():
    out_names = ["var_0", "var_1"]
    channel_dim = -3
    area = torch.tensor([])  # area not used by this config
    mapping_loss_config = WeightedMappingLossConfig(
        type="MSE", weights={"var_0": 4.0, "var_1": 1.0}
    )
    normalizer = StandardNormalizer(
        means={name: torch.as_tensor(0.0) for name in out_names},
        stds={name: torch.as_tensor(1.0) for name in out_names},
    )

    gridded_operations: GriddedOperations = LatLonOperations(area)

    mapping_loss = mapping_loss_config.build(
        gridded_operations,
        out_names=out_names,
        channel_dim=channel_dim,
        normalizer=normalizer,
    )

    x0 = torch.ones(4, 5, 5).to(get_device())
    x1 = 2.0 * x0
    y0 = 2.0 * x0
    y1 = 2.5 * x0

    x_mapping = {"var_0": x0, "var_1": x1}
    y_mapping = {"var_0": y0, "var_1": y1}

    assert mapping_loss(x_mapping, y_mapping) == ((16 + 0.25) / 2.0)


@pytest.mark.parametrize("mean", [0.0, 1.0])
@pytest.mark.parametrize("scale", [1.0, 2.0])
def test_WeightedMappingLoss_with_ensemble_dim(mean, scale):
    loss = torch.nn.MSELoss()
    n_channels = 5
    n_ensemble = 3
    packer = Packer([f"var_{i}" for i in range(n_channels)])
    out_names = [f"var_{i}" for i in range(n_channels)]
    normalizer = StandardNormalizer(
        means={name: torch.as_tensor(mean) for name in out_names},
        stds={name: torch.as_tensor(scale) for name in out_names},
    )
    mapping_loss = WeightedMappingLoss(
        loss,
        weights={},
        out_names=out_names,
        normalizer=normalizer,
    )
    x = torch.randn(
        15,
        n_ensemble,
        n_channels,
        10,
        10,
    ).to(get_device(), dtype=torch.float)
    y = torch.randn(
        15,
        1,  # only 1 ensemble member for target
        n_channels,
        10,
        10,
    ).to(get_device(), dtype=torch.float)
    x_mapping = {name: x[:, :, i, :, :] for i, name in enumerate(packer.names)}
    y_mapping = {name: y[:, :, i, :, :] for i, name in enumerate(packer.names)}
    assert torch.allclose(mapping_loss(x_mapping, y_mapping), loss(x, y) / scale**2)


def test_WeightedMappingLoss_with_target_nans():
    loss = torch.nn.MSELoss()
    n_channels = 5
    packer = Packer([f"var_{i}" for i in range(n_channels)])
    out_names = [f"var_{i}" for i in range(n_channels)]
    normalizer = StandardNormalizer(
        means={name: torch.as_tensor(0.0) for name in out_names},
        stds={name: torch.as_tensor(1.0) for name in out_names},
    )
    mapping_loss = WeightedMappingLoss(
        loss,
        weights={},
        out_names=out_names,
        normalizer=normalizer,
    )
    x = torch.randn(
        15,
        n_channels,
        10,
        10,
    ).to(get_device(), dtype=torch.float)
    y = torch.randn(
        15,
        n_channels,
        10,
        10,
    ).to(get_device(), dtype=torch.float)
    x_mapping = {name: x[:, i, :, :].clone() for i, name in enumerate(packer.names)}
    y_mapping = {name: y[:, i, :, :].clone() for i, name in enumerate(packer.names)}
    y_mapping[packer.names[0]][:, :, 0] = float("nan")
    x[:, 0, :, 0] = 0.0
    y[:, 0, :, 0] = 0.0
    assert torch.allclose(mapping_loss(x_mapping, y_mapping), loss(x, y))



import abc
import dataclasses
import warnings
from collections.abc import Callable

# we use Type to distinguish from type attr of StepSelector
from typing import Any, ClassVar, Type, TypeVar, cast, final  # noqa: UP035

import torch
from torch import nn

from fme.core.dataset_info import DatasetInfo
from fme.core.normalizer import StandardNormalizer
from fme.core.ocean import OceanConfig
from fme.core.registry.registry import Registry
from fme.core.typing_ import TensorDict, TensorMapping


# Children still need to decorate with @dataclass, otherwise
# they will be a dataclass with no dataclass fields.
@dataclasses.dataclass
class StepConfigABC(abc.ABC):
    @abc.abstractmethod
    def get_step(
        self,
        dataset_info: DatasetInfo,
        init_weights: Callable[[list[nn.Module]], None],
    ) -> "StepABC":
        """
        Args:
            dataset_info: Information about the training dataset.
            init_weights: Function to initialize the weights of the step before
                wrapping in DistributedDataParallel. This is particularly useful
                when freezing parameters, as the DistributedDataParallel will
                otherwise expect frozen weights to have gradients, and will
                raise an exception.

        Returns:
            The state of the stepper.
        """

    @property
    @abc.abstractmethod
    def n_ic_timesteps(self) -> int:
        pass

    @property
    @abc.abstractmethod
    def input_names(self) -> list[str]:
        pass

    @property
    @abc.abstractmethod
    def output_names(self) -> list[str]:
        """
        Names of variables output by the step.
        """
        pass

    @property
    @abc.abstractmethod
    def next_step_input_names(self) -> list[str]:
        """
        Names of variables required in next_step_input_data for .step.
        """
        pass

    @property
    @final
    def prognostic_names(self) -> list[str]:
        return list(set(self.input_names).intersection(self.output_names))

    @property
    @abc.abstractmethod
    def loss_names(self) -> list[str]:
        """
        Names of variables to be included in the loss function.
        """
        pass

    @abc.abstractmethod
    def get_next_step_forcing_names(self) -> list[str]:
        pass

    @abc.abstractmethod
    def get_loss_normalizer(
        self,
        extra_names: list[str] | None = None,
        extra_residual_scaled_names: list[str] | None = None,
    ) -> StandardNormalizer:
        """
        Args:
            extra_names: Names of additional variables to include in the
                loss normalizer.
            extra_residual_scaled_names: extra_names which use residual scale factors,
                if enabled.

        Returns:
            The loss normalizer.
        """

    @abc.abstractmethod
    def replace_ocean(self, ocean: OceanConfig | None):
        pass

    @abc.abstractmethod
    def get_ocean(self) -> OceanConfig | None:
        pass

    @abc.abstractmethod
    def load(self):
        """
        Update configuration in-place so it does not depend on external files.
        """
        pass


T = TypeVar("T", bound=StepConfigABC)


@dataclasses.dataclass
class StepSelector(StepConfigABC):
    type: str
    config: dict[str, Any]
    registry: ClassVar[Registry] = Registry()

    def __post_init__(self):
        self._step_config_instance: StepConfigABC = cast(
            StepConfigABC, self.registry.get(self.type, self.config)
        )

    @property
    def n_ic_timesteps(self) -> int:
        return self._step_config_instance.n_ic_timesteps

    @classmethod
    def register(cls, name: str) -> Callable[[Type[T]], Type[T]]:  # noqa: UP006
        return cls.registry.register(name)

    def get_step(
        self,
        dataset_info: DatasetInfo,
        init_weights: Callable[[list[nn.Module]], None] = lambda x: None,
    ) -> "StepABC":
        """
        Args:
            dataset_info: Information about the training dataset.
            init_weights: Function to initialize the weights of the step before
                wrapping in DistributedDataParallel. This is particularly useful
                when freezing parameters, as the DistributedDataParallel will
                otherwise expect frozen weights to have gradients, and will
                raise an exception.

        Returns:
            The state of the stepper.
        """
        return self._step_config_instance.get_step(dataset_info, init_weights)

    @classmethod
    def get_available_types(cls) -> set[str]:
        """This class method is used to expose all available types of Steps."""
        return set(cls(type="", config={}).registry._types.keys())

    def get_next_step_forcing_names(self) -> list[str]:
        return self._step_config_instance.get_next_step_forcing_names()

    @property
    def input_names(self) -> list[str]:
        return self._step_config_instance.input_names

    @property
    def output_names(self) -> list[str]:
        """
        Names of variables output by the step.
        """
        return self._step_config_instance.output_names

    @property
    def next_step_input_names(self) -> list[str]:
        """
        Names of variables required in next_step_input_data for .step.
        """
        return self._step_config_instance.next_step_input_names

    @property
    def loss_names(self) -> list[str]:
        """
        Names of variables to be included in the loss function.
        """
        return self._step_config_instance.loss_names

    def get_loss_normalizer(
        self,
        extra_names: list[str] | None = None,
        extra_residual_scaled_names: list[str] | None = None,
    ) -> StandardNormalizer:
        return self._step_config_instance.get_loss_normalizer(
            extra_names=extra_names,
            extra_residual_scaled_names=extra_residual_scaled_names,
        )

    def replace_ocean(self, ocean: OceanConfig | None):
        self._step_config_instance.replace_ocean(ocean)
        self.config = dataclasses.asdict(self._step_config_instance)

    def get_ocean(self) -> OceanConfig | None:
        return self._step_config_instance.get_ocean()

    def load(self):
        self._step_config_instance.load()
        self.config = dataclasses.asdict(self._step_config_instance)


class StepABC(abc.ABC, nn.Module):
    SelfType = TypeVar("SelfType", bound="StepABC")

    @property
    @abc.abstractmethod
    def config(self) -> StepConfigABC:
        pass

    @final
    def get_loss_normalizer(
        self,
        extra_names: list[str] | None = None,
        extra_residual_scaled_names: list[str] | None = None,
    ) -> StandardNormalizer:
        return self.config.get_loss_normalizer(
            extra_names=extra_names,
            extra_residual_scaled_names=extra_residual_scaled_names,
        )

    @property
    @final
    def n_ic_timesteps(self) -> int:
        return self.config.n_ic_timesteps

    @property
    @final
    def input_names(self) -> list[str]:
        return self.config.input_names

    @property
    @final
    def output_names(self) -> list[str]:
        return self.config.output_names

    @property
    @final
    def prognostic_names(self) -> list[str]:
        return self.config.prognostic_names

    @property
    @final
    def loss_names(self) -> list[str]:
        return self.config.loss_names

    @property
    @abc.abstractmethod
    def modules(self) -> nn.ModuleList:
        pass

    @property
    @abc.abstractmethod
    def normalizer(self) -> StandardNormalizer:
        pass

    @property
    @final
    def next_step_input_names(self) -> list[str]:
        """
        Names of variables required in next_step_input_data for .step.
        """
        return self.config.next_step_input_names

    @property
    @final
    def next_step_forcing_names(self) -> list[str]:
        """Names of input variables which come from the output timestep."""
        return self.config.get_next_step_forcing_names()

    @property
    @abc.abstractmethod
    def surface_temperature_name(self) -> str | None:
        """
        Name of the surface temperature variable, if one is available.
        """
        pass

    @property
    @abc.abstractmethod
    def ocean_fraction_name(self) -> str | None:
        """
        Name of the ocean fraction variable, if one is available.
        """
        pass

    @abc.abstractmethod
    def get_regularizer_loss(self) -> torch.Tensor:
        """
        Get the regularizer loss.
        """
        pass

    @abc.abstractmethod
    def step(
        self: SelfType,
        input: TensorMapping,
        next_step_input_data: TensorMapping,
        wrapper: Callable[[nn.Module], nn.Module] = lambda x: x,
    ) -> TensorDict:
        """
        Step the model forward one timestep given input data.

        Args:
            input: Mapping from variable name to tensor of shape
                [n_batch, n_lat, n_lon]. This data is used as input for pytorch
                module(s) and is assumed to contain all input variables
                and be denormalized.
            next_step_input_data: Mapping from variable name to tensor of shape
                [n_batch, n_lat, n_lon]. This must contain the necessary input
                data at the output timestep, such as might be needed to prescribe
                sea surface temperature or use a corrector.
            wrapper: Wrapper to apply over each nn.Module before calling.

        Returns:
            The denormalized output data at the next time step.
        """
        pass

    @final
    def forward(
        self, input: TensorMapping, next_step_input_data: TensorMapping
    ) -> TensorDict:
        return self.step(input, next_step_input_data)

    @final
    def export(
        self: SelfType,
        input: TensorMapping,
        next_step_input_data: TensorMapping,
    ) -> torch.export.ExportedProgram:
        """
        Script the step function.
        """
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message=".*does not reference an nn.Module.*"
            )
            return torch.export.export(self, (input, next_step_input_data))

    @abc.abstractmethod
    def get_state(self) -> dict[str, Any]:
        """
        Returns:
            The state of the step object as expected by load_state,
                may or may not include initialization parameters.
        """
        pass

    @abc.abstractmethod
    def load_state(self, state: dict[str, Any]):
        """
        Load the state of the step object.
        """
        pass



import contextlib

import torch

USE_CPU_RANDN = False


def randn_like(x: torch.Tensor, **kwargs):
    if USE_CPU_RANDN:
        device = kwargs.pop("device", x.device)
        return torch.randn_like(x, device="cpu", **kwargs).to(device)
    else:
        return torch.randn_like(x, **kwargs)


def randn(shape: torch.Size, **kwargs):
    if USE_CPU_RANDN:
        device = kwargs.pop("device", None)
        return torch.randn(shape, device="cpu", **kwargs).to(device)
    else:
        return torch.randn(shape, **kwargs)


@contextlib.contextmanager
def use_cpu_randn():
    """
    Context manager to use CPU when generating random numbers for
    randn and randn_like.

    This is likely less performant than generating them directly on the GPU,
    but it allows comparing regression outputs between machines.
    """
    global USE_CPU_RANDN
    old_use_cpu_randn = USE_CPU_RANDN
    USE_CPU_RANDN = True
    yield
    USE_CPU_RANDN = old_use_cpu_randn



import abc

from fme.core.typing_ import TensorDict, TensorMapping


class CorrectorABC(abc.ABC):
    @abc.abstractmethod
    def __call__(
        self,
        input_data: TensorMapping,
        gen_data: TensorMapping,
        forcing_data: TensorMapping,
    ) -> TensorDict: ...
